"""Interact with the a posix file system."""

from __future__ import annotations

import pathlib
from typing import AsyncIterator, Sequence, Union

from metadata_crawler._helper import posix as _posix_rs

from ..api.storage_backend import MetadataType, PathTemplate


class PosixPath(PathTemplate):
    """Class to interact with a Posix file system."""

    _fs_type = "posix"

    async def is_dir(self, path: Union[str, pathlib.Path]) -> bool:
        """Check if a given path is a directory object on the storage system.

        Parameter
        ---------
        path : str, pathlib.Path
            Path of the object store

        Returns
        -------
        bool: True if path is dir object, False if otherwise or doesn't exist
        """
        return await _posix_rs.is_dir(str(path))

    async def is_file(self, path: Union[str, pathlib.Path]) -> bool:
        """Check if a given path is a file object on the storage system.

        Parameter
        ---------
        path : str, pathlib.Path
            Path of the object store


        Returns
        -------
        bool: True if path is file object, False if otherwise or doesn't exist
        """
        return await _posix_rs.is_file(str(path))

    async def iterdir(self, path: Union[str, pathlib.Path]) -> AsyncIterator[str]:
        """Get all sub directories from a given path.

        Parameter
        ---------
        path : str, pathlib.Path
            Path of the object store

        Yields
        ------
        str: 1st level sub directory
        """
        entries: Sequence[str] = await _posix_rs.iterdir(str(path))
        for entry in entries:
            yield entry

    async def rglob(
        self, path: Union[str, pathlib.Path], glob_pattern: str = "*"
    ) -> AsyncIterator[MetadataType]:
        """Search recursively for paths matching a given glob pattern.

        Parameter
        ---------
        path : str, pathlib.Path
            Path of the object store
        glob_pattern: str
            Pattern that the target files must match

        Yields
        ------
        MetadataType: Path of the object store that matches the glob pattern.
        """
        for p in await _posix_rs.rglob(
            str(path),
            glob_pattern,
            list(self.suffixes) or None,
        ):
            yield MetadataType(path=p, metadata={})

    def path(self, path: Union[str, pathlib.Path]) -> str:
        """Get the full path (including any schemas/netlocs).

        Parameters
        ----------
        path: str, pathlib.Path
            Path of the object store

        Returns
        -------
        str:
            URI of the object store
        """
        return str(pathlib.Path(path).absolute())

    def uri(self, path: Union[str, pathlib.Path]) -> str:
        """Get the uri of the object store.

        Parameters
        ----------
        path: str, pathlib.Path
            Path of the object store

        Returns
        -------
        str:
            URI of the object store
        """
        return f"file://{pathlib.Path(path).absolute()}"



"""Interact with the OpenStack swift cloud."""

from __future__ import annotations

import asyncio
import pathlib
from fnmatch import fnmatch
from typing import AsyncIterator, Dict, List, Optional, Tuple, Union, cast
from urllib.parse import SplitResult, urljoin, urlsplit, urlunparse

import aiohttp
import fsspec

from ..api.storage_backend import MetadataType, PathTemplate


def _basename(key: str) -> str:
    return pathlib.PosixPath(key[:-1] if key.endswith("/") else key).name


class SwiftPath(PathTemplate):
    """Class to interact with the OpenStack swift cloud storage system."""

    _fs_type = "swift"

    def __post_init__(self) -> None:
        self.storage_options = self.storage_options or {}
        self.os_password = self.storage_options.get("os_password", self._pw)
        self.os_user_id = self.storage_options.get("os_user_id", self._user)
        self.os_project_id = self.storage_options.get("os_project_id")
        self.os_auth_token = self.storage_options.get("os_auth_token") or None
        self._os_storage_url = self.storage_options.get(
            "os_storage_url", ""
        ).rstrip("/")
        self.os_auth_url = self.storage_options.get(
            "os_auth_url", self._guess_tempauth_url(self._os_storage_url)
        )
        self._container = self.storage_options.get(
            "container", self._os_storage_url.split("/")[-1]
        ).rstrip("/")
        self._os_storage_url = self._os_storage_url.removesuffix(self._container)
        self._url_split: Optional[SplitResult] = None

    @staticmethod
    def _guess_tempauth_url(storage_url: str) -> str:
        """Construct the swift url.

        Heuristic: For TempAuth, switch '/v1/...' to '/auth/v1.0' on same host:port.
        Returns None if storage_url doesn't look like a Swift v1 endpoint.
        """
        p = urlsplit(storage_url)
        # Typical Swift proxy paths: '/v1/...' or '/swift/v1/...'
        if not (p.path.startswith("/v1/") or p.path.startswith("/swift/v1/")):
            return ""
        # Use same scheme+netloc, set path to /auth/v1.0
        return urlunparse((p.scheme, p.netloc, "/auth/v1.0", "", "", ""))

    @property
    def storage_path(self) -> str:
        """Path part of the storage url."""
        split = self.url_split
        return "/" + split.path.lstrip("/").rstrip("/")

    @property
    def url_split(self) -> SplitResult:
        """Retrieve the split parts of the storage url."""
        if self._url_split is not None:
            return self._url_split
        if not self._os_storage_url:
            raise RuntimeError("os_storage_url must be set")
        storage_url = self._os_storage_url.removesuffix(self._container)
        self._url_split = urlsplit(urljoin(storage_url, self._container))
        return self._url_split

    @property
    def _anon(self) -> bool:
        """Decide if we can logon at all."""
        return False if self.os_password or self.headers else True

    async def logon(self) -> None:
        """Logon to the swfit system if necessary."""
        headers = {
            "X-Auth-User": f"{self.os_project_id}:{self.os_user_id}",
            "X-Auth-Key": self.os_password,
        }
        async with aiohttp.ClientSession() as session:
            async with session.get(self.os_auth_url, headers=headers) as res:
                if res.status != 200:
                    raise ValueError(f"Logon to {self.os_auth_url} failed")
                self.os_auth_token = res.headers["X-Auth-Token"]

    def _is_zarr_like_match(self, key: str, glob_pattern: str) -> bool:
        key_l = key.lower()
        base = _basename(key)
        if key_l.endswith(".zarr") or key_l.endswith(".zarr/"):
            if ".zarr" in self.suffixes and fnmatch(base, glob_pattern):
                return True
        return False

    async def _url_fragments(self, url: str) -> Tuple[str, str]:
        url_split = urlsplit(url)
        url_path = (
            ("/" + url_split.path.lstrip("/"))
            .removeprefix(self.storage_path)
            .rstrip("/")
            .lstrip("/")
        )

        parsed_url = SplitResult(
            url_split.scheme or self.url_split.scheme,
            url_split.netloc or self.url_split.netloc,
            f"{self.storage_path}/{url_path}",
            url_split.query,
            url_split.fragment,
        )
        _path = pathlib.PosixPath(parsed_url.path).parts[1:]
        url_prefix = "/".join(_path[:3])
        prefix = "/".join(_path[3:])
        if prefix:
            prefix += "/"
        url_head = f"{parsed_url.scheme}://{parsed_url.netloc}/{url_prefix}"
        return url_head, prefix

    async def _read_json(
        self, path: str, delimiter: Optional[str] = "/"
    ) -> List[Dict[str, str]]:
        url, prefix = await self._url_fragments(path)
        suffix = f"?format=json&prefix={prefix}"
        if delimiter:
            suffix += f"&delimiter={delimiter}"
        else:
            suffix = suffix.rstrip("/")
        url = f"{url}{suffix}"
        errors = {
            403: PermissionError(f"Permission denied for {path}"),
            404: FileNotFoundError(f"No such file or directory {path}"),
        }
        async with aiohttp.ClientSession() as session:
            for _ in range(2):
                async with session.get(url, headers=self.headers) as res:
                    if res.status < 300:
                        return cast(list[dict[str, str]], await res.json())
                    if res.status == 401:
                        await self.logon()
                        continue
        raise errors.get(res.status, RuntimeError(f"Failed to query {path}"))

    def _get_dir_from_path(self, data: dict[str, str]) -> str | None:
        if (
            data.get("subdir")
            or data.get("content_type", "") == "application/directory"
        ):
            return data.get("subdir") or data.get("name")
        return None

    @property
    def headers(self) -> dict[str, str]:
        """Define the headers used to interact with swift."""
        if self.os_auth_token is None:
            return {}
        return {"X-Auth-Token": self.os_auth_token}

    async def is_file(self, path: Union[str, pathlib.Path]) -> bool:
        """Check if a given path is a file object on the storage system."""
        try:
            data = (await self._read_json(str(path)))[0]
        except (FileNotFoundError, IndexError):
            return False
        return self._get_dir_from_path(data) is None

    async def is_dir(self, path: Union[str, pathlib.Path]) -> bool:
        """Check if a given path is a directory object on the storage system."""
        try:
            data = (await self._read_json(str(path)))[0]
        except (FileNotFoundError, IndexError):
            return False
        return self._get_dir_from_path(data) is not None

    async def iterdir(self, path: Union[str, pathlib.Path]) -> AsyncIterator[str]:
        """Get all sub directories of a directory."""
        try:
            for data in await self._read_json(str(path)):
                new_path = self._get_dir_from_path(data)
                if new_path:
                    out = (
                        str(path).lstrip("/")
                        + "/"
                        + pathlib.PosixPath(new_path).name
                    )
                    yield out
        except (FileNotFoundError, PermissionError):
            pass

    async def rglob(
        self,
        path: Union[str, pathlib.Path],
        glob_pattern: str = "*",
    ) -> AsyncIterator[MetadataType]:
        """Search recursively for files matching a glob_pattern."""
        delimiter: Optional[str] = None
        if await self.is_dir(path):
            delimiter = "/"
        for data in await self._read_json(str(path), delimiter=delimiter):
            # swift doesn't natively support pagination, so we need to do it
            # ourselves.
            name = data.get("name")
            dir_name = self._get_dir_from_path(data)
            if dir_name:
                # if it's an actual object named foo.zarr, treat as zarr store
                if self._is_zarr_like_match(dir_name, glob_pattern):
                    yield MetadataType(path=dir_name.rstrip("/"), metadata={})
                else:
                    async for md in self.rglob(dir_name, glob_pattern):
                        yield md
            elif name:
                if pathlib.PosixPath(name).suffix in self.suffixes and fnmatch(
                    name, glob_pattern
                ):
                    yield MetadataType(path=name, metadata={})

    def get_fs_and_path(self, uri: str) -> Tuple[fsspec.AbstractFileSystem, str]:
        """Return (fs, path) suitable for xarray.

        Parameters
        ----------
        uri:
            Path to the object store / file name


        Returns
        -------
        fsspec.AbstractFileSystem, str:
            The AbstractFileSystem class and the corresponding path to the
            data store.
        """
        url_split = urlsplit(uri)
        url_path = (
            ("/" + url_split.path.lstrip("/"))
            .removeprefix(self.storage_path)
            .rstrip("/")
            .lstrip("/")
        )
        url = SplitResult(
            url_split.scheme or self.url_split.scheme,
            url_split.netloc or self.url_split.netloc,
            f"{self.storage_path}/{url_path}",
            url_split.query,
            url_split.fragment,
        ).geturl()
        if not self._anon:
            asyncio.run(self.logon())
        return (
            fsspec.filesystem(
                "http",
                headers=self.headers,
                block_size=2**20,
            ),
            url,
        )

    def path(self, path: Union[str, pathlib.Path]) -> str:
        """Get the full path (including any schemas/netlocs).

        Parameters
        ----------
        path: str, asyncio.Path, pathlib.Path
            Path of the object store

        Returns
        -------
        str:
            URI of the object store
        """
        url_split = urlsplit(str(path))
        if not url_split.netloc:
            path = f"{self.url_split.path}/{url_split.path}"
        else:
            path = url_split.path

        res = SplitResult(
            url_split.scheme or self.url_split.scheme,
            url_split.netloc or self.url_split.netloc,
            path,
            url_split.query,
            url_split.fragment,
        ).geturl()
        return res

    def uri(self, path: Union[str, pathlib.Path]) -> str:
        """Get the uri of the object store.

        Parameters
        ----------
        path: str, asyncio.Path, pathlib.Path
            Path of the object store

        Returns
        -------
        str:
            URI of the object store
        """
        return self.path(path)



"""Metadata Crawler API high level functions."""

import asyncio
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import tomlkit

try:
    import uvloop

    use_uvloop = True
except ImportError:

    use_uvloop = False  # pragma: no cover


from ._version import __version__
from .api.config import ConfigMerger, DRSConfig
from .api.metadata_stores import CatalogueBackendType, IndexName
from .data_collector import DataCollector
from .logger import logger
from .run import async_add, async_delete, async_index

if use_uvloop:
    async_model = uvloop
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
else:
    async_model = asyncio  # pragma: no cover

__all__ = [
    "logger",
    "__version__",
    "DataCollector",
    "index",
    "add",
    "delete",
    "get_config",
    "async_index",
    "async_delete",
    "async_add",
    "get_config",
]


def get_config(*config: Union[Path, str]) -> ConfigMerger:
    """Get a drs config file merged with the default config.

    The method is helpful to inspect all possible configurations and their
    default values.

    Parameters
    ^^^^^^^^^^

    config:
        Path to a user defined config file that is going to be merged with
        the default config.
    """
    _ = DRSConfig.load(*config)
    return ConfigMerger(*config)


def index(
    index_system: str,
    *catalogue_files: Union[Path, str, List[str], List[Path]],
    batch_size: int = 2500,
    verbosity: int = 0,
    log_suffix: Optional[str] = None,
    **kwargs: Any,
) -> None:
    """Index metadata in the indexing system.

    Parameters
    ^^^^^^^^^^

    index_system:
        The index server where the metadata is indexed.
    catalogue_files:
        Path to the file(s) where the metadata was stored.
    batch_size:
        If the index system supports batch-sizes, the size of the batches.
    verbosity:
        Set the verbosity level.
    log_suffix:
        Add a suffix to the log file output.

    Other Parameters
    ^^^^^^^^^^^^^^^^

    **kwargs:
        Keyword arguments used to delete data from the index.

    Examples
    ^^^^^^^^

    .. code-block:: python

        index(
            "solr",
            "/tmp/catalog-1.yml",
            "/tmp/catalog-2.yml",
            batch_size=50,
            server="localhost:8983",
        )
    """
    async_model.run(
        async_index(
            index_system,
            *catalogue_files,
            batch_size=batch_size,
            verbosity=verbosity,
            log_suffix=log_suffix,
            **kwargs,
        )
    )


def delete(
    index_system: str,
    batch_size: int = 2500,
    verbosity: int = 0,
    log_suffix: Optional[str] = None,
    **kwargs: Any,
) -> None:
    """Delete metadata from the indexing system.

    Parameters
    ^^^^^^^^^^

    index_system:
        The index server where the metadata is indexed.
    batch_size:
        If the index system supports batch-sizes, the size of the batches.
    verbosity:
        Set the verbosity of the system.
    log_suffix:
        Add a suffix to the log file output.

    Other Parameters
    ^^^^^^^^^^^^^^^^

    **kwargs:
        Keyword arguments used to delete data from the index.


    Examples
    ^^^^^^^^

    .. code-block:: python

        delete(
            "solr",
            server="localhost:8983",
            facets=[("project", "CMIP6"), ("institute", "MPI-M")],
        )
    """
    async_model.run(
        async_delete(
            index_system, batch_size=batch_size, log_suffix=log_suffix, **kwargs
        )
    )


def add(
    *config_files: Union[Path, str, Dict[str, Any], tomlkit.TOMLDocument],
    store: Optional[Union[str, Path]] = None,
    data_object: Optional[Union[str, List[str]]] = None,
    data_set: Optional[Union[str, List[str]]] = None,
    data_store_prefix: str = "metadata",
    catalogue_backend: CatalogueBackendType = "jsonlines",
    batch_size: int = 25_000,
    comp_level: int = 4,
    storage_options: Optional[Dict[str, Any]] = None,
    shadow: Optional[Union[str, List[str]]] = None,
    latest_version: str = IndexName().latest,
    all_versions: str = IndexName().all,
    n_procs: Optional[int] = None,
    verbosity: int = 0,
    log_suffix: Optional[str] = None,
    password: bool = False,
    fail_under: int = -1,
    **kwargs: Any,
) -> None:
    """Harvest metadata from storage systems and add them to an intake catalogue.

    .. versionchanged:: 2511.0.0

       The catalogue argument has been rearanged and is now a keyword
       argument: ``add("data.yaml", "drs-config.toml")`` becomes
       ``add("drs-config.toml", store="data.yaml")``. If the ``store`` keyword
       is omitted the output catalogue will be interpreted as config file.

    Parameters
    ^^^^^^^^^^

    config_files:
        Path to the drs-config file / loaded configuration.
    store:
        Path to the intake catalogue where the collected metadata will be
        stored.
    data_ojbect:
        Instead of defining datasets that are to be crawled you can crawl
        data based on their directories. The directories must be a root dirs
        given in the drs-config file. By default all root dirs are crawled.
    data_set:
        Datasets that should be crawled. The datasets need to be defined
        in the drs-config file. By default all datasets are crawled.
        Names can contain wildcards such as ``xces-*``.
    data_store_prefix:
        Absolute path or relative path to intake catalogue source
    data_dir:
        Instead of defining datasets are are to be crawled you can crawl
        data based on their directories. The directories must be a root dirs
        given in the drs-config file. By default all root dirs are crawled.
    bach_size:
        Batch size that is used to collect the meta data. This can affect
        performance.
    comp_level:
        Compression level used to write the meta data to csv.gz
    storage_options:
        Set additional storage options for adding metadata to the metadata store
    shadow:
        'Shadow' this storage options. This is useful to hide secrets in public
        data catalogues.
    catalogue_backend:
        Intake catalogue backend
    latest_version:
        Name of the core holding 'latest' metadata.
    all_versions:
        Name of the core holding 'all' metadata versions.
    password:
        Display a password prompt and set password before beginning.
    n_procs:
        Set the number of parallel processes for collecting.
    verbosity:
        Set the verbosity of the system.
    log_suffix:
        Add a suffix to the log file output.
    fail_under:
         Fail if less than X of the discovered files could be indexed.

    Other Parameters
    ^^^^^^^^^^^^^^^^

    **kwargs:
        Additional keyword arguments.


    Examples
    ^^^^^^^^

    .. code-block:: python

        add(
            "~/data/drs-config.toml",
            store="my-data.yaml",
            data_set=["cmip6", "cordex"],
        )
    """
    async_model.run(
        async_add(
            *config_files,
            store=store,
            data_object=data_object,
            data_set=data_set,
            batch_size=batch_size,
            comp_level=comp_level,
            password=password,
            catalogue_backend=catalogue_backend,
            data_store_prefix=data_store_prefix,
            shadow=shadow,
            latest_version=latest_version,
            all_versions=all_versions,
            n_procs=n_procs,
            storage_options=storage_options,
            verbosity=verbosity,
            log_suffix=log_suffix,
            fail_under=fail_under,
            **kwargs,
        )
    )



"""Special unit tests for the LookupMixin class."""

import importlib
import os
import time


def _reload_with_cache_dir(tmp_path):
    os.environ["MDC_LOOKUP_CACHE_DIR"] = str(tmp_path)
    import metadata_crawler.api.mixin.lookup_mixin as lm  # noqa

    return importlib.reload(lm)


def test_static_cmor_fastpath(monkeypatch, tmp_path):
    lm = _reload_with_cache_dir(tmp_path)

    class Fake(lm.LookupMixin):
        calls = 0

        def read_attr(self, attribute: str, path: str, **k):
            Fake.calls += 1
            return f"{attribute}@{path}"

    # provide a static entry; read_attr must NOT be called
    lm.LookupMixin.CMOR_STATIC = {("tier",): "STATIC"}

    f = Fake()
    out = f.lookup("/p/file.nc", "attr", "tier")
    assert out == "STATIC"
    assert Fake.calls == 0

    # clean up cache
    lm._DC.clear()


def test_diskcache_reuse_across_instances(monkeypatch, tmp_path):
    lm = _reload_with_cache_dir(tmp_path)

    class Fake(lm.LookupMixin):
        calls = 0

        def read_attr(self, attribute: str, path: str, **k):
            Fake.calls += 1
            # tiny sleep to make any races more likely (still deterministic serially)
            time.sleep(0.01)
            return f"{attribute}@{path}"

    # ensure no static entries interfere
    lm.LookupMixin.CMOR_STATIC = {}

    a = Fake()
    b = Fake()

    # first call: miss -> compute -> add
    v1 = a.lookup("/p/file.nc", "units", "cmip6", "day", "tas")
    assert v1 == "units@/p/file.nc"
    assert Fake.calls == 1

    # second call (new instance): hit from diskcache, no new compute
    v2 = b.lookup("/p/file.nc", "units", "cmip6", "day", "tas")
    assert v2 == v1
    assert Fake.calls == 1  # unchanged

    # clean up cache
    lm._DC.clear()



"""Random utility functions."""

import difflib
import logging
import multiprocessing as mp
import multiprocessing.context as mctx
import os
import sys
import time
from datetime import datetime, timedelta
from importlib.metadata import entry_points
from typing import (
    IO,
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Protocol,
    Set,
    Tuple,
    TypeAlias,
    TypeVar,
    Union,
)

import ciso8601
import orjson
import rich.console
import rich.spinner
from dateutil.parser import isoparse
from rich.live import Live
from rich.progress import Progress, TaskID

from ..logger import logger

T = TypeVar("T")
U = TypeVar("U")


class SimpleQueueLike(Protocol[T]):
    """A simple queue like Type class."""

    def put(self, item: T) -> None:  # noqa
        ...

    def get(self) -> T:  # noqa
        ...


class QueueLike(Protocol[T]):
    """A queue like Type class."""

    def put(self, item: T) -> None:  # noqa
        ...

    def get(
        self, block: bool = True, timeout: Optional[float] = ...
    ) -> T:  # noqa
        ...

    def qsize(self) -> int:  # noqa
        ...


class EventLike(Protocol):
    """An event like Type class."""

    def set(self) -> None:  # noqa
        ...

    def clear(self) -> None:  # noqa
        ...

    def is_set(self) -> bool:  # noqa
        ...

    def wait(self) -> None:  # noqa
        ...


class LockLike(Protocol):
    """A lock like Type class."""

    def acquire(
        self, blocking: bool = ..., timeout: Optional[float] = ...
    ) -> bool:  # noqa
        ...

    def release(self) -> None:  # noqa
        ...

    def __enter__(self) -> "LockLike": ...
    def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None: ...


class ValueLike(Protocol[U]):
    """A value like Type class."""

    value: U

    def get_lock(self) -> "Any":  # noqa
        ...


class FilesystemLike(Protocol):
    """File-like opener protocol (e.g., fsspec)."""

    def open(
        self,
        path: str,
        mode: str = "rt",
        compression: Optional[str] = None,
        encoding: Optional[str] = None,
        **kwargs: Any,
    ) -> IO[str]:  # noqa
        ...


Counter: TypeAlias = ValueLike[int]
PrintLock = mp.Lock()
Console = rich.console.Console(force_terminal=sys.stdout.isatty(), stderr=True)


class MetadataCrawlerException(Exception):
    """Custom Exception for the crawling."""


class EmptyCrawl(MetadataCrawlerException):
    """Cusotom Exceptoin for a crawl with no results."""


async def create_async_iterator(itt: Iterable[Any]) -> AsyncIterator[Any]:
    """Create an async iterator from as sync iterable."""
    for item in itt:
        yield item


def _parse_iso_datetime(s: str) -> datetime:
    return ciso8601.parse_datetime(s) or datetime.fromisoformat(s)


def parse_batch(
    lines: List[str],
    timestamp_keys: Set[str],
) -> List[Dict[str, Any]]:
    """Parse a batch of NDJSON lines and convert timestamp fields.

    Parameters
    ^^^^^^^^^^
    lines : list of str
        Raw NDJSON lines.
    timestamp_keys : set of str
        Keys that should be parsed as datetimes.

    Returns
    ^^^^^^^
    list of dict
        Parsed objects with timestamp fields converted to ``datetime``.
    """
    out: List[Dict[str, Any]] = []
    append = out.append
    loads = orjson.loads
    parse_dt = _parse_iso_datetime

    for line in lines:
        obj: Dict[str, Any] = loads(line)
        for k in timestamp_keys:
            v = obj.get(k, None)
            if v is None:
                continue
            if isinstance(v, str):
                obj[k] = parse_dt(v)
            elif isinstance(v, list):
                obj[k] = [parse_dt(x) if isinstance(x, str) else x for x in v]
        append(obj)
    return out


def convert_str_to_timestamp(
    time_str: str, alternative: str = "0001-01-01"
) -> datetime:
    """Convert a string representation of a time step to an iso timestamp.

    Parameters
    ----------
    time_str: str
        Representation of the time step in formats:
        - %Y%m%d%H%M%S%f (year, month, day, hour, minute, second, millisecond)
        - %Y%m%d%H%M (year, month, day, hour, minute)
        - %Y%m (year, month)
        - %Y%m%dT%H%M (year, month, day, hour, minute with T separator)
        - %Y%j (year and day of year, e.g. 2022203 for 22nd July 2022)
        - %Y (year only)
    alternative: str, default: 0
        If conversion fails, the alternative/default value the time step
        gets assign to

    Returns
    -------
    str: ISO time string representation of the input time step, such as
        %Y %Y-%m-%d or %Y-%m-%dT%H%M%S
    """
    _date = isoparse(alternative)
    _time = f"{_date.strftime('%H')}:{_date.strftime('%M')}"
    _day = _date.strftime("%d")
    _mon = _date.strftime("%m")
    has_t_separator = "T" in time_str
    position_t = time_str.find("T") if has_t_separator else -1
    # Strip anything that's not a number from the string
    if not time_str:
        return _date
    # Not valid if time repr empty or starts with a letter, such as 'fx'
    digits = "".join(filter(str.isdigit, time_str))
    l_times = len(digits)
    if not l_times:
        return _date
    try:
        if l_times <= 4:
            # Suppose this is a year only
            return isoparse(f"{digits.zfill(4)}-{_mon}-{_day}T{_time}")
        if l_times <= 6:
            # Suppose this is %Y%m or %Y%e
            return isoparse(f"{digits[:4]}-{digits[4:].zfill(2)}-{_day}T{_time}")
        # Year and day of year
        if l_times == 7:
            # Suppose this is %Y%j
            year = int(digits[:4])
            day_of_year = int(digits[4:])
            date = datetime(year, 1, 1, _date.hour, _date.minute) + timedelta(
                days=day_of_year - 1
            )
            return date
        if l_times <= 8:
            # Suppose this is %Y%m%d
            return isoparse(
                f"{digits[:4]}-{digits[4:6]}-{digits[6:].zfill(2)}T{_time}"
            )

        date_str = f"{digits[:4]}-{digits[4:6]}-{digits[6:8]}"
        time = digits[8:]
        if len(time) <= 2:
            time = time.zfill(2)
        else:
            # Alaways drop seconds
            time = time[:2] + ":" + time[2 : min(4, len(time))].zfill(2)
        return isoparse(f"{date_str}T{time}")

    except ValueError:
        if has_t_separator and position_t > 0:
            date_part = time_str[:position_t]
            time_part = time_str[position_t + 1 :]

            date_digits = "".join(filter(str.isdigit, date_part))
            if len(date_digits) >= 8:
                return isoparse(
                    f"{date_digits[:4]}-{date_digits[4:6]}"
                    f"-{date_digits[6:8]}T{time_part[:2].zfill(2)}"
                )

        return _date


def find_closest(msg: str, target: str, options: Iterable[str]) -> str:
    """Find the closest match for a target within a collection of items.

    Parameters
    ----------
    target:   The string to match.
    options:  A list of candidate strings.


    Returns
    -------
        str: Message
    """
    matches = difflib.get_close_matches(target, options, n=1, cutoff=0.6)
    suffix = f", did you mean {matches[0]}?" if matches else ""
    return msg + suffix


def load_plugins(group: str) -> Dict[str, Any]:
    """Load harverster plugins."""
    eps = entry_points().select(group=group)
    plugins = {}
    for ep in eps:
        plugins[ep.name] = ep.load()
    return plugins


def exception_handler(exception: BaseException) -> None:
    """Handle raising exceptions appropriately."""
    msg = str(exception)
    if logger.level >= logging.INFO:
        msg += " - increase verbosity for more information"
        exc_info = None
    else:
        exc_info = exception
    logger.critical(msg, exc_info=exc_info)
    raise SystemExit(1)


def daemon(
    func: Callable[..., Any],
) -> Callable[..., mctx.ForkProcess]:
    """Threading decorator.

    use @daemon above the function you want to run in the background
    """

    def background_func(*args: Any, **kwargs: Any) -> mctx.ForkProcess:
        ctx = mp.get_context("fork")
        proc = ctx.Process(target=func, args=args, kwargs=kwargs, daemon=True)
        proc.start()
        return proc

    return background_func


def timedelta_to_str(seconds: Union[int, float]) -> str:
    """Convert seconds to a more human readable format."""
    hours = seconds // 60**2
    minutes = (seconds // 60) % 60
    sec = round(seconds - (hours * 60 + minutes) * 60, 2)
    out = []
    for num, letter in {sec: "Sec.", minutes: "Min.", hours: "Hour"}.items():
        if num > 0:
            out.append(f"{num} {letter}")
    return " ".join(out[::-1])


class IndexProgress:
    """A helper that displays the progress of index Tasks."""

    def __init__(
        self,
        total: int = 0,
        interactive: Optional[bool] = None,
        text: str = "Indexing: ",
    ) -> None:
        if interactive is None:
            self._interactive = bool(
                int(os.getenv("MDC_INTERACTIVE", str(int(Console.is_terminal))))
            )
        else:
            self._interactive = interactive
        self._log_interval = int(os.getenv("MDC_LOG_INTERVAL", "30"))
        self.text = text
        self._done = 0
        self._task: TaskID = TaskID(0)
        self._total = total
        self._start = self._last_log = time.time()
        self._progress = Progress()
        self._last_printed_percent: float = -1.0

    def start(self) -> None:
        """Start the progress bar."""
        self._start = self._last_log = time.time()

        if self._interactive:
            self._task = self._progress.add_task(
                f"[green] {self.text}", total=self._total or None
            )
            self._progress.start()

    def stop(self) -> None:
        """Stop the progress bar."""
        if self._interactive:
            self._progress.stop()
        else:
            self._text_update()

    def _text_update(self, bar_width: int = 40) -> None:
        elapsed = timedelta(seconds=int(time.time() - self._start))
        log_interval = timedelta(seconds=int(time.time() - self._last_log))
        if self._total > 0:
            filled = int((self._last_printed_percent / 100) * bar_width)
            bar = "#" * filled + "-" * (bar_width - filled)
            text = f"{self.text} [{bar}] {self._last_printed_percent:>6,.02f}%"
        else:
            text = f"{self.text} [{self._done:>12,}]"
        if log_interval.total_seconds() >= self._log_interval:
            print(f"{text} ({elapsed})", flush=True)
            self._last_log = time.time()

    def update(self, inc: int) -> None:
        """Update the status progress bar by an increment."""
        self._done += inc

        if self._interactive is True:
            desc = f"{self.text} [{self._done:>10d}]" if self._done == 0 else None
            self._progress.update(self._task, advance=inc, description=desc)
            return

        frac = self._done / max(self._total, 1)
        pct = frac * 100
        if pct > self._last_printed_percent or self._total == 0:
            self._last_printed_percent = pct
            self._text_update()


@daemon
def print_performance(
    print_status: EventLike,
    num_files: Counter,
    ingest_queue: QueueLike[Any],
    num_objects: Counter,
) -> None:
    """Display the progress of the crawler."""
    spinner = rich.spinner.Spinner(
        os.getenv("SPINNER", "earth"), text="[b]Preparing crawler ...[/]"
    )
    interactive = bool(
        int(os.getenv("MDC_INTERACTIVE", str(int(Console.is_terminal))))
    )
    log_interval = int(os.getenv("MDC_LOG_INTERVAL", "30"))
    sample_interval = 1.0 if interactive else 10.0

    def _snapshot() -> Tuple[float, int, int, int]:
        start = time.monotonic()
        n0 = num_files.value
        time.sleep(sample_interval)
        dn = num_files.value - n0
        dt = max(1e-6, time.monotonic() - start)
        perf_file = dn / dt
        queue_size = ingest_queue.qsize()
        return perf_file, n0, queue_size, num_objects.value

    def _build_msg(
        perf_file: float,
        discovered: int,
        queue_size: int,
        indexed: int,
        *,
        markup: bool,
    ) -> str:
        # Color thresholds only when markup=True (interactive)
        if markup:
            f_col = (
                "green"
                if perf_file > 500
                else "red" if perf_file < 100 else "blue"
            )
            q_col = (
                "red"
                if queue_size > 100_000
                else "green" if queue_size < 10_000 else "blue"
            )
            return (
                f"[bold]Discovering: [{f_col}]{perf_file:>6,.1f}[/{f_col}] files/s "
                f"#files: [blue]{discovered:>10,.0f}[/blue] "
                f"in queue: [{q_col}]{queue_size:>6,.0f}[/{q_col}] "
                f"#indexed: [blue]{indexed:>10,.0f}[/blue][/bold]"
            )
        else:
            return (
                f"Discovering: {perf_file:,.1f} files/s | "
                f"files={discovered:,} | queue={queue_size:,} | indexed={indexed:,}"
            )

    if interactive:
        with Live(
            spinner, console=Console, refresh_per_second=2.5, transient=True
        ):
            while print_status.is_set():
                perf, disc, qsz, idx = _snapshot()
                spinner.update(text=_build_msg(perf, disc, qsz, idx, markup=True))
        # Clear the last line when done
        Console.print(" " * Console.width, end="\r")
        Console.print(" ")
    else:
        # Non-TTY (e.g. systemd): emit a plain summary every log_interval secs
        next_log = time.monotonic()
        while print_status.is_set():
            perf, disc, qsz, idx = _snapshot()
            now = time.monotonic()
            if now >= next_log:
                # Print one clean line; journald/Cockpit will show one entry
                print(_build_msg(perf, disc, qsz, idx, markup=False), flush=True)
                next_log = now + log_interval



"""Setup for the tests."""

import multiprocessing as mp
from pathlib import Path
from queue import Queue
from tempfile import TemporaryDirectory
from threading import Thread
from typing import Any, Dict, Iterator

import numpy as np
import pandas as pd
import pytest
import requests
import toml
import xarray as xr
from pymongo import MongoClient


class ThreadContext:
    """Fake the mp.get_context with threads."""

    def SimpleQueue(self) -> Queue[Any]:
        """Alias for Queue."""
        return Queue()

    def Queue(self) -> Queue[Any]:
        """Alias for Queue."""
        return Queue()

    def Process(self, *args: Any, **kwargs: Any) -> Thread:
        """Alias for Thread."""
        return Thread(*args, **kwargs)

    def Value(slef, *args: Any, **kwargs: Any) -> Any:
        return mp.Value("i", 0)


@pytest.fixture(autouse=True)
def mock_subprocess(monkeypatch) -> Iterator[None]:
    """Multiprocess -> Thread."""
    import metadata_crawler.api.metadata_stores as stores_mod
    import metadata_crawler.utils as md_utils

    monkeypatch.setattr(
        stores_mod.mp,
        "get_context",
        lambda method="spawn": ThreadContext(),
        raising=True,
    )
    monkeypatch.setattr(
        md_utils.mp,
        "get_context",
        lambda method="spawn": ThreadContext(),
        raising=True,
    )
    yield


@pytest.fixture(scope="function")
def metadata() -> Dict[str, Any]:
    return [
        {
            "file": "/foo1.nc",
            "uri": "file:///foo1.nc",
            "project": ["CMIP6"],
            "product": ["CMIP"],
            "institute": None,
            "model": ["ACCESS-CM2"],
            "experiment": ["amip"],
            "time_frequency": ["mon"],
            "realm": ["atmos"],
            "cmor_table": ["Amon"],
            "ensemble": ["r1i1p1f1"],
            "variable": ["ua"],
            "grid_label": ["gn"],
            "version": "v20191108",
            "driving_model": None,
            "rcm_name": None,
            "rcm_version": None,
            "dataset": "cmip6-fs",
            "format": "nc",
            "grid_id": None,
            "level_type": ["2d"],
            "user": None,
            "fs_type": "posix",
            "time_aggregation": ["mean"],
        },
        {
            "file": "/foo2.nc",
            "uri": "file:///foo2.nc",
            "project": ["CMIP6"],
            "product": ["CMIP"],
            "institute": None,
            "model": ["MPI-ESM1-2-LR"],
            "experiment": ["amip"],
            "time_frequency": ["mon"],
            "realm": ["atmos"],
            "cmor_table": ["Amon"],
            "ensemble": ["r2i1p1f1"],
            "variable": ["ua"],
            "grid_label": ["gn"],
            "version": "v20190815",
            "driving_model": None,
            "rcm_name": None,
            "rcm_version": None,
            "dataset": "cmip6-fs",
            "format": "nc",
            "grid_id": None,
            "level_type": ["2d"],
            "user": None,
            "fs_type": "posix",
            "time_aggregation": ["mean"],
        },
    ]


@pytest.fixture(scope="session")
def data_dir() -> Iterator[Path]:
    """Define the directory where the data is stored."""
    this_dir = Path(__file__).parent
    yield (this_dir.parent / "data").absolute()


@pytest.fixture(scope="function")
def mongo_server() -> Iterator[Dict[str, str]]:
    server = "mongodb://mongo:secret@localhost:27017"
    db = "metadata"
    with MongoClient(server) as client:
        _db = client[db]
        for col in ("latest", "files"):
            _db[col].delete_many({})
    try:
        yield {"url": server, "database": db}
    finally:
        with MongoClient(server) as client:
            db = client[db]
            for col in ("latest", "files"):
                db[col].delete_many({})


@pytest.fixture(scope="function")
def solr_server() -> Iterator[str]:
    solr_server = "http://localhost:8983"
    for core in ("files", "latest"):
        res = requests.post(
            f"{solr_server}/solr/{core}/update/json?commit=true",
            json={"delete": {"query": "*:*"}},
        )
        res.raise_for_status()
    try:
        yield solr_server
    finally:
        for core in ("files", "latest"):
            res = requests.post(
                f"{solr_server}/solr/{core}/update/json?commit=true",
                json={"delete": {"query": "*:*"}},
            )
            res.raise_for_status()


@pytest.fixture(scope="session")
def storage_options() -> Dict[str, str]:
    """Define s3 connection options."""
    return {
        "endpoint_url": "http://localhost:9000",
        "key": "minioadmin",
        "secret": "minioadmin",
    }


@pytest.fixture(scope="session")
def drs_config_path(data_dir: Path) -> Iterator[Path]:
    """Define the main config path."""
    with TemporaryDirectory() as temp_dir:
        for _file in "drs_config.toml", "benchmark-config.toml":
            config_file = data_dir.parent / _file
            assert config_file.is_file()  # Smoke test
            config = toml.loads(config_file.read_text())
            for key, cfg in config.items():
                if not hasattr(cfg, "get"):
                    continue
                if cfg.get("root_path") and ":" not in cfg["root_path"]:
                    p = Path(cfg["root_path"])
                    if not p.is_absolute():
                        cfg["root_path"] = str((data_dir.parent / p).absolute())

            temp_path = Path(temp_dir) / _file
            with temp_path.open("w") as stream:
                toml.dump(config, stream)
        yield Path(temp_dir) / "drs_config.toml"


@pytest.fixture(scope="function")
def cat_file() -> Iterator[Path]:
    with TemporaryDirectory() as temp_dir:
        yield Path(temp_dir) / "data.yml"


@pytest.fixture(scope="session")
def dataset() -> Iterator[xr.Dataset]:
    lon = xr.DataArray(
        np.linspace(-120, -30, 20),
        name="lon",
        dims=("lon",),
        coords={
            "long_name": "Longitude",
            "short_name": "lon",
            "units": "degress_east",
            "axis": "X",
        },
    )
    lat = xr.DataArray(
        np.linspace(-40, 10, 50),
        name="lat",
        dims=("lat",),
        coords={
            "long_name": "Latitude",
            "short_name": "lat",
            "units": "degress_north",
            "axis": "Y",
        },
    )
    time = xr.DataArray(
        pd.date_range("2020-01-01", "2025-12-31", freq="1YE"),
        name="time",
        dims=("time",),
        attrs={"long_name": "Time", "short_name": "time", "axis": "T"},
    )
    tas = xr.DataArray(
        np.random.rayleigh(size=(len(time), len(lat), len(lon))),
        name="tas",
        attrs={
            "long_name": "2M Air temperature",
            "short_name": "tas",
            "units": "degC",
            "member": "foo",
            "grid_label": "gn",
        },
        coords={"time": time, "lat": lat, "lon": lon},
    )
    yield xr.Dataset(
        {"tas": tas},
        attrs={
            "product": "foo",
            "frequency": "1yr",
            "institute_id": "bar",
            "experiment": "random",
        },
    )


@pytest.fixture(scope="session")
def zarr_data(dataset: xr.Dataset) -> Iterator[Path]:
    with TemporaryDirectory() as temp_dir:
        out = Path(temp_dir) / "tas.zarr"
        dataset.to_zarr(out)
        yield out



"""API for adding commands to the cli."""

from functools import wraps
from typing import Any, Callable, Dict, Tuple, Union

from pydantic import BaseModel, ConfigDict


class Parameter(BaseModel):
    """CLI Parameter model."""

    model_config = ConfigDict(extra="allow")

    args: Union[str, Tuple[str, ...]]
    """Names for the arpargse.Namespace."""
    help: str
    """Help string that is going to be displayed."""


def cli_parameter(*args: str, **kwargs: Any) -> Dict[str, Any]:
    """Construct a ``argparse.Namespace``.

    Parameters
    ^^^^^^^^^^
    *args:
        Any arguments passed to ``argparse.ArgumentParser().add_argument``
    **kwargs:
        Any keyword arguments passed to ``argparse.ArgumentParser().add_arguent``

    """
    return Parameter(args=args, **kwargs).model_dump()


def cli_function(
    help: str = "",
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Wrap command line arguments around a method.

    Those arguments represent the arguments you would normally use to create
    a `argparse subcommand <https://docs.python.org/3/library/argparse.html>`_.

    Parameters
    ^^^^^^^^^^
    help:
        Help string for this sub command.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        setattr(func, "_cli_help", help or func.__doc__)

        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            return func(*args, **kwargs)

        return wrapper

    return decorator



"""Metadata Storage definitions."""

from __future__ import annotations

import abc
import asyncio
import gzip
import json
import multiprocessing as mp
import os
import re
import time
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from enum import Enum
from io import BytesIO
from multiprocessing import sharedctypes
from pathlib import Path
from types import NoneType
from typing import (
    Any,
    AsyncIterator,
    ClassVar,
    Dict,
    List,
    Literal,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Type,
    TypeAlias,
    Union,
    cast,
)

import fsspec
import orjson
import yaml

import metadata_crawler

from ..logger import logger
from ..utils import (
    Counter,
    MetadataCrawlerException,
    QueueLike,
    SimpleQueueLike,
    create_async_iterator,
    parse_batch,
)
from .config import DRSConfig, SchemaField
from .storage_backend import MetadataType

ISO_FORMAT_REGEX = re.compile(
    r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:\.\d+)?Z?$"
)

BATCH_SECS_THRESHOLD = 20
BATCH_ITEM = List[Tuple[str, Dict[str, Any]]]


ConsumerQueueType: TypeAlias = QueueLike[
    Union[int, Tuple[str, str, MetadataType]]
]
WriterQueueType: TypeAlias = SimpleQueueLike[Union[int, BATCH_ITEM]]


class Stream(NamedTuple):
    """A representation of a path stream as named tuple."""

    name: str
    path: str


class DateTimeEncoder(json.JSONEncoder):
    """JSON‐Encoder that emits datetimes as ISO‐8601 strings."""

    def default(self, obj: Any) -> Any:
        """Set default time encoding."""
        if isinstance(obj, datetime):
            return obj.isoformat()
        return super().default(obj)


class DateTimeDecoder(json.JSONDecoder):
    """JSON Decoder that converts ISO‐8601 strings to datetime objects."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(object_hook=self._decode_objects, *args, **kwargs)

    def _decode_datetime(self, obj: Any) -> Any:
        if isinstance(obj, list):
            return list(map(self._decode_datetime, obj))
        elif isinstance(obj, dict):
            for key in obj:
                obj[key] = self._decode_datetime(obj[key])
        if isinstance(obj, str):
            try:
                return datetime.fromisoformat(obj.replace("Z", "+00:00"))
            except ValueError:
                return obj
        return obj

    def _decode_objects(self, obj: Dict[str, Any]) -> Any:
        for key, value in obj.items():
            obj[key] = self._decode_datetime(value)
        return obj


class IndexName(NamedTuple):
    """A paired set of metadata indexes representations.

        - `latest`: Metadata for the latest version of each dataset.
        - `files`: Metadata for all available versions of datasets.

    This abstraction is backend-agnostic and can be used with any index system,
    such as Apache Solr cores, MongoDB collections, or SQL tables.

    """

    latest: str = "latest"
    all: str = "files"


class IndexStore:
    """Base class for all metadata stores."""

    suffix: ClassVar[str]
    """Path suffix of the metadata store."""

    driver: ClassVar[str]
    """Intake driver."""

    def __init__(
        self,
        path: str,
        index_name: IndexName,
        schema: Dict[str, SchemaField],
        batch_size: int = 25_000,
        mode: Literal["r", "w"] = "r",
        storage_options: Optional[Dict[str, Any]] = None,
        shadow: Optional[Union[str, List[str]]] = None,
        **kwargs: Any,
    ) -> None:
        self.storage_options = storage_options or {}
        self._shadow_options = (
            shadow or [] if isinstance(shadow, (list, NoneType)) else [shadow]
        )
        self._ctx = mp.get_context("spawn")
        self.queue: WriterQueueType = self._ctx.SimpleQueue()
        self._sent = 42
        self._fs, self._is_local_path = self.get_fs(path, **self.storage_options)
        self._path = self._fs.unstrip_protocol(path)
        self.schema = schema
        self.batch_size = batch_size
        self.index_names: Tuple[str, str] = (index_name.latest, index_name.all)
        self.mode = mode
        self._rows_since_flush = 0
        self._last_flush = time.time()
        self._paths: List[Stream] = []
        self.max_workers: int = max(1, (os.cpu_count() or 4))
        for name in self.index_names:
            out_path = self.get_path(name)
            self._paths.append(Stream(name=name, path=out_path))
        self._timestamp_keys: Set[str] = {
            k
            for k, col in schema.items()
            if getattr(getattr(col, "base_type", None), "value", None)
            == "timestamp"
        }

    @staticmethod
    def get_fs(
        uri: str, **storage_options: Any
    ) -> Tuple[fsspec.AbstractFileSystem, bool]:
        """Get the base-url from a path."""
        protocol, path = fsspec.core.split_protocol(uri)
        protocol = protocol or "file"
        add = {"anon": True} if protocol == "s3" else {}
        storage_options = storage_options or add
        fs = fsspec.filesystem(protocol, **storage_options)
        return fs, protocol == "file"

    @abc.abstractmethod
    async def read(
        self,
        index_name: str,
    ) -> AsyncIterator[List[Dict[str, Any]]]:
        """Yield batches of metadata records from a specific table.

        Parameters
        ^^^^^^^^^^
        index_name:
            The name of the index_name.

        Yields
        ^^^^^^
        List[Dict[str, Any]]:
            Deserialised metadata records.
        """
        yield [{}]  # pragma: no cover

    def get_path(self, path_suffix: Optional[str] = None) -> str:
        """Construct a path name for a given suffix."""
        path = self._path.removesuffix(self.suffix)
        new_path = (
            f"{path}-{path_suffix}{self.suffix}"
            if path_suffix
            else f"{path}{self.suffix}"
        )
        return new_path

    def join(self) -> None:
        """Shutdown the writer task."""
        self.queue.put(self._sent)
        if self.proc is not None:
            self.proc.join()

    def close(self) -> None:
        """Shutdown the write worker."""
        self.join()

    @property
    def proc(self) -> Optional["mp.process.BaseProcess"]:
        """The writer process."""
        raise NotImplementedError("This property must be defined.")

    @abc.abstractmethod
    def get_args(self, index_name: str) -> Dict[str, Any]:
        """Define the intake arguments."""
        ...  # pragma: no cover

    def catalogue_storage_options(
        self, path: Optional[str] = None
    ) -> Dict[str, Any]:
        """Construct the storage options for the catalogue."""
        is_s3 = (path or "").startswith("s3://")
        opts = {
            k: v
            for k, v in self.storage_options.items()
            if k not in self._shadow_options
        }
        shadow_keys = {
            "key",
            "secret",
            "token",
            "username",
            "user",
            "password",
            "secret_file",
            "secretfile",
        }
        opts |= {"anon": True} if is_s3 and not shadow_keys & opts.keys() else {}
        return opts


class JSONLineWriter:
    """Write JSONLines to disk."""

    def __init__(
        self,
        *streams: Stream,
        comp_level: int = 4,
        shadow: Optional[Union[str, List[str]]] = None,
        **storage_options: Any,
    ) -> None:

        self._comp_level = comp_level
        self._f: Dict[str, BytesIO] = {}
        self._streams = {s.name: s.path for s in streams}
        self._records = 0
        self.storage_options = storage_options
        for _stream in streams:
            fs, _ = IndexStore.get_fs(_stream.path, **storage_options)
            parent = os.path.dirname(_stream.path).rstrip("/")
            try:
                fs.makedirs(parent, exist_ok=True)
            except Exception:  # pragma: no cover
                pass  # pragma: no cover
            self._f[_stream.name] = fs.open(_stream.path, mode="wb")

    @classmethod
    def as_daemon(
        cls,
        queue: WriterQueueType,
        semaphore: int,
        *streams: Stream,
        comp_level: int = 4,
        **storage_options: Any,
    ) -> None:
        """Start the writer process as a daemon."""
        this = cls(*streams, comp_level=comp_level, **storage_options)
        get = queue.get
        add = this._add
        while True:
            item = get()
            if item == semaphore:
                logger.info("Closing writer task.")
                break
            try:
                add(cast(BATCH_ITEM, item))
            except Exception as error:
                logger.error(error)
        this.close()

    @staticmethod
    def _encode_records(records: List[Dict[str, Any]]) -> bytes:
        """Serialize a list of dicts into one JSONL bytes blob."""
        parts = [orjson.dumps(rec) for rec in records]
        return b"".join(p + b"\n" for p in parts)

    def _gzip_once(self, payload: bytes) -> bytes:
        """Compress a whole JSONL blob into a single gz member (fast)."""
        return gzip.compress(payload, compresslevel=self._comp_level)

    def _add(self, metadata_batch: List[Tuple[str, Dict[str, Any]]]) -> None:
        """Add a batch of metadata to the gzip store."""
        by_index: Dict[str, List[Dict[str, Any]]] = {
            name: [] for name in self._streams
        }
        for index_name, metadata in metadata_batch:
            by_index[index_name].append(metadata)
        for index_name, records in by_index.items():
            if not records:
                continue
            payload = self._encode_records(records)
            gz = self._gzip_once(payload)
            self._f[index_name].write(gz)
            self._records += len(records)

    def close(self) -> None:
        """Close the files."""
        for name, stream in self._f.items():
            try:
                stream.flush()
            except Exception:
                pass
            stream.close()
            if not self._records:
                fs, _ = IndexStore.get_fs(
                    self._streams[name], **self.storage_options
                )
                fs.rm(self._streams[name])


class JSONLines(IndexStore):
    """Write metadata to gzipped JSONLines files."""

    suffix = ".json.gz"
    driver = "intake.source.jsonfiles.JSONLinesFileSource"

    def __init__(
        self,
        path: str,
        index_name: IndexName,
        schema: Dict[str, SchemaField],
        mode: Literal["w", "r"] = "r",
        storage_options: Optional[Dict[str, Any]] = None,
        shadow: Optional[Union[str, List[str]]] = None,
        batch_size: int = 25_000,
        **kwargs: Any,
    ):
        super().__init__(
            path,
            index_name,
            schema,
            mode=mode,
            shadow=shadow,
            storage_options=storage_options,
            batch_size=batch_size,
        )
        _comp_level = int(kwargs.get("comp_level", "4"))
        self._proc: Optional["mp.process.BaseProcess"] = None
        if mode == "w":
            self._proc = self._ctx.Process(
                target=JSONLineWriter.as_daemon,
                args=(
                    self.queue,
                    self._sent,
                )
                + tuple(self._paths),
                kwargs={**{"comp_level": _comp_level}, **self.storage_options},
                daemon=True,
            )
            self._proc.start()

    @property
    def proc(self) -> Optional["mp.process.BaseProcess"]:
        """The writer process."""
        return self._proc

    def get_args(self, index_name: str) -> Dict[str, Any]:
        """Define the intake arguments."""
        path = self.get_path(index_name)
        return {
            "urlpath": path,
            "compression": "gzip",
            "text_mode": True,
            "storage_options": self.catalogue_storage_options(path),
        }

    async def read(
        self,
        index_name: str,
    ) -> AsyncIterator[List[Dict[str, Any]]]:
        """Yield batches of metadata records from a specific table.

        Parameters
        ^^^^^^^^^^
        index_name:
            The name of the index_name.

        Yields
        ^^^^^^^
        List[Dict[str, Any]]:
            Deserialised metadata records.
        """
        loop = asyncio.get_running_loop()
        ts_keys = self._timestamp_keys
        path = self.get_path(index_name)
        with (
            self._fs.open(
                path,
                mode="rt",
                compression="gzip",
                encoding="utf-8",
            ) as stream,
            ThreadPoolExecutor(max_workers=self.max_workers) as pool,
        ):
            raw_lines: List[str] = []
            async for line in create_async_iterator(stream):
                raw_lines.append(line)
                if len(raw_lines) >= self.batch_size:
                    batch = await loop.run_in_executor(
                        pool, parse_batch, raw_lines, ts_keys
                    )
                    yield batch
                    raw_lines.clear()
            if raw_lines:
                batch = await loop.run_in_executor(
                    pool, parse_batch, raw_lines, ts_keys
                )
                yield batch


class CatalogueBackends(Enum):
    """Define the implemented catalogue backends."""

    jsonlines = JSONLines


CatalogueBackendType: TypeAlias = Literal["jsonlines"]


class CatalogueReader:
    """Backend for reading the content of an intake catalogue.

    Parameters
    ^^^^^^^^^^
    catalogue_file:
        Path to the intake catalogue
    batch_size:
        Size of the metadata chunks that should be read.
    """

    def __init__(
        self,
        catalogue_file: Union[str, Path],
        batch_size: int = 2500,
        storage_options: Optional[Dict[str, Any]] = None,
    ) -> None:
        catalogue_file = str(catalogue_file)
        storage_options = storage_options or {}
        cat = self.load_catalogue(catalogue_file, **storage_options)
        _schema_json = cat["metadata"]["schema"]
        schema = {s["key"]: SchemaField(**s) for k, s in _schema_json.items()}
        index_name = IndexName(**cat["metadata"]["index_names"])
        cls: Type[IndexStore] = CatalogueBackends[
            cat["metadata"]["backend"]
        ].value
        storage_options = cat["metadata"].get("storage_options", {})
        self.store = cls(
            cat["metadata"]["prefix"],
            index_name,
            schema,
            mode="r",
            batch_size=batch_size,
            storage_options=storage_options,
        )

    @staticmethod
    def load_catalogue(path: Union[str, Path], **storage_options: Any) -> Any:
        """Load a intake yaml catalogue (remote or local)."""
        fs, _ = IndexStore.get_fs(str(path), **storage_options)
        cat_path = fs.unstrip_protocol(path)
        with fs.open(cat_path) as stream:
            return yaml.safe_load(stream.read())


class QueueConsumer:
    """Class that consumes the file discovery queue."""

    def __init__(
        self,
        config: Dict[str, Any],
        num_objects: "sharedctypes.Synchronized[Any]",
        writer_queue: WriterQueueType,
        silent: bool = False,
    ) -> None:
        self.config = DRSConfig(**config)
        self._writer_queue = writer_queue
        self.num_objects = num_objects
        self._silent = silent

    def _flush_batch(
        self,
        batch: List[Tuple[str, Dict[str, Any]]],
    ) -> None:
        logger.info("Ingesting %i items", len(batch))
        try:
            self._writer_queue.put(batch.copy())
            if self._silent is False:
                with self.num_objects.get_lock():
                    self.num_objects.value += len(batch)
        except Exception as error:  # pragma: no cover
            logger.error(error)  # pragma: no cover
        batch.clear()

    @classmethod
    def run_consumer_task(
        cls,
        queue: ConsumerQueueType,
        writer_queue: WriterQueueType,
        config: Dict[str, Any],
        num_objects: "sharedctypes.Synchronized[Any]",
        batch_size: int,
        poison_pill: int,
        silent: bool = False,
    ) -> None:
        """Set up a consumer task waiting for incoming data to be ingested."""
        this = cls(config, num_objects, writer_queue, silent=silent)
        this_worker = os.getpid()
        logger.info("Adding %i consumer to consumers.", this_worker)
        batch: List[Tuple[str, Dict[str, Any]]] = []
        append = batch.append
        read_metadata = this.config.read_metadata
        flush = this._flush_batch
        get = queue.get
        while True:
            item = get()
            if item == poison_pill:
                break
            try:
                name, drs_type, inp = cast(Tuple[str, str, MetadataType], item)
                metadata = read_metadata(drs_type, inp)
            except MetadataCrawlerException as error:
                logger.warning(error)
                continue
            except Exception as error:
                logger.error(error)
                continue
            append((name, metadata))
            if len(batch) >= batch_size:
                flush(batch)
        if batch:
            flush(batch)
        logger.info("Closing consumer %i", this_worker)


class CatalogueWriter:
    """Create intake catalogues that store metadata entries.

    Parameters
    ^^^^^^^^^^
    yaml_path:
        Path the to intake catalogue that should be created.
    index_name:
        Names of the metadata indexes.
    config:
        Metadata Config class
    data_store_prefix:
        Prefix of the path/url where the metadata is stored.
    batch_size:
        Size of the metadata chunks that should be added to the data store.
    n_procs:
        Number of processes collecting metadata
    storage_options:
        Set additional storage options for adding metadata to the metadata store
    shadow:
        'Shadow' this storage options. This is useful to hide secrets in public
        data catalogues.
    """

    def __init__(
        self,
        yaml_path: str,
        index_name: IndexName,
        config: DRSConfig,
        *,
        data_store_prefix: str = "metadata",
        backend: str = "jsonlines",
        batch_size: int = 25_000,
        n_procs: Optional[int] = None,
        storage_options: Optional[Dict[str, Any]] = None,
        shadow: Optional[Union[str, List[str]]] = None,
        **kwargs: Any,
    ) -> None:
        self.config = config
        storage_options = storage_options or {}
        self.fs, _ = IndexStore.get_fs(yaml_path, **storage_options)
        self.path = self.fs.unstrip_protocol(yaml_path)
        scheme, _, _ = data_store_prefix.rpartition("://")
        self.silent = bool(int(os.getenv("MDC_SILENT", "0")))
        self.backend = backend
        if not scheme and not os.path.isabs(data_store_prefix):
            data_store_prefix = os.path.join(
                os.path.abspath(os.path.dirname(yaml_path)), data_store_prefix
            )
        self.prefix = data_store_prefix
        self.index_name = index_name
        cls: Type[IndexStore] = CatalogueBackends[backend].value
        self.store = cls(
            data_store_prefix,
            index_name,
            self.config.index_schema,
            mode="w",
            storage_options=storage_options,
            shadow=shadow,
            **kwargs,
        )
        self._ctx = mp.get_context("spawn")
        self.queue: ConsumerQueueType = self._ctx.Queue()
        self._poison_pill = 13
        self.num_objects: Counter = self._ctx.Value("i", 0)
        n_procs = n_procs or min(mp.cpu_count(), 15)
        batch_size_per_proc = max(int(batch_size / n_procs), 100)
        self._tasks = [
            self._ctx.Process(
                target=QueueConsumer.run_consumer_task,
                args=(
                    self.queue,
                    self.store.queue,
                    getattr(self.config, "_model_dict", {}),
                    self.num_objects,
                    batch_size_per_proc,
                    self._poison_pill,
                ),
                kwargs={"silent": self.silent},
            )
            for i in range(n_procs)
        ]

    async def put(
        self,
        inp: MetadataType,
        drs_type: str,
        name: str = "",
    ) -> None:
        """Add items to the fifo queue.

        This method is used by the data crawling (discovery) method
        to add the name of the catalogue, the path to the input file object
        and a reference of the Data Reference Syntax class for this
        type of dataset.

        Parameters
        ^^^^^^^^^^
        inp:
            Path and metadata of the discovered object.
        drs_type:
            The data type the discovered object belongs to.
        name:
            Name of the catalogue, if applicable. This variable depends on
            the cataloguing system. For example apache solr would use a `core`.
        """
        self.queue.put((name, drs_type, inp))

    @property
    def ingested_objects(self) -> int:
        """Get the number of ingested objects."""
        return self.num_objects.value

    @property
    def size(self) -> int:
        """Get the size of the worker queue."""
        return self.queue.qsize()

    def join_all_tasks(self) -> None:
        """Block the execution until all tasks are marked as done."""
        logger.debug("Releasing consumers from their duty.")
        for _ in self._tasks:
            self.queue.put(self._poison_pill)
        for task in self._tasks:
            task.join()
        self.store.join()

    async def close(self, create_catalogue: bool = True) -> None:
        """Close any connections."""
        self.store.join()
        self.store.close()
        if create_catalogue:
            self._create_catalogue_file()

    async def delete(self) -> None:
        """Delete all stores."""
        await self.close(False)
        for name in self.index_name.latest, self.index_name.all:
            path = self.store.get_path(name)
            self.store._fs.rm(path) if self.store._fs.exists(path) else None
        self.fs.rm(self.path) if self.fs.exists(self.path) else None

    def run_consumer(self) -> None:
        """Set up all the consumers."""
        for task in self._tasks:
            task.start()

    def _create_catalogue_file(self) -> None:
        catalog = {
            "description": (
                f"{metadata_crawler.__name__} "
                f"(v{metadata_crawler.__version__})"
                f" at {datetime.now().strftime('%c')}"
            ),
            "metadata": {
                "version": 1,
                "backend": self.backend,
                "prefix": self.prefix,
                "storage_options": self.store.catalogue_storage_options(
                    self.prefix
                ),
                "index_names": {
                    "latest": self.index_name.latest,
                    "all": self.index_name.all,
                },
                "indexed_objects": self.ingested_objects,
                "schema": {
                    k: json.loads(s.model_dump_json())
                    for k, s in self.store.schema.items()
                },
            },
            "sources": {
                self.index_name.latest: {
                    "description": "Latest metadata versions.",
                    "driver": self.store.driver,
                    "args": self.store.get_args(self.index_name.latest),
                },
                self.index_name.all: {
                    "description": "All metadata versions only.",
                    "driver": self.store.driver,
                    "args": self.store.get_args(self.index_name.all),
                },
            },
        }
        with self.fs.open(self.path, "w", encoding="utf-8") as f:
            yaml.safe_dump(
                catalog,
                f,
                sort_keys=False,  # preserve our ordering
                default_flow_style=False,
            )



"""Testing the config module."""

import json
import os
from datetime import datetime
from pathlib import Path
from tempfile import TemporaryDirectory

import intake
import mock
import pytest
import toml
import xarray as xr
from jinja2 import Template

from metadata_crawler import add, index
from metadata_crawler.api.config import DRSConfig
from metadata_crawler.api.metadata_stores import DateTimeDecoder, DateTimeEncoder
from metadata_crawler.backends.posix import PosixPath
from metadata_crawler.backends.swift import SwiftPath
from metadata_crawler.utils import MetadataCrawlerException

CONFIG = """[bar]
root_path = "{{path | default('/foo')}}"
drs_format = "bar"
fs_type = "{{ fs_type | default('posix') }}"
[bar.defaults]
project = "observations"
[drs_settings.dialect.bar]
sources    = {{ sources | default(["data"]) }}
inherits_from = "{{inherits | default('freva')}}"
[drs_settings.dialect.bar.data_specs.read_kws]
engine = "zarr"
[drs_settings.dialect.bar.data_specs.globals]
institute = "institution_id"
experiment = "experiment_id"
ensemble = "ensemble"
grid_label = "grid_label"
data_vars = "__variable__"
[drs_settings.dialect.bar.data_specs.var_attrs]
variable = {{vars}}
ensemble = {var = "tas", attr = "member"}
grid_id = {var = "tas", attr = "grid_id", defualt = "foo"}
[drs_settings.dialect.bar.data_specs.stats]
bbox = {stat = "bbox", coords = ["lat", "lon"]}
time = {stat = "minmax", var = "time"}
min = {stat = "min", var = "tas" }
max = {stat = "max", var = "tas" }
nan = {stat = "min", var = "foo", default = 0}
freq = {stat = "timedelta", var = "time"}
time_range = {stat = "range", coords = "time" }
[drs_settings.dialect.bar.special.model]
type = "call"
call = "float({% raw %}{{ max }}{% endraw %})-float({% raw %}{{ min }}{% endraw %})"
[drs_settings.storage_options]
anon = true
[drs_settings.defaults]
cmor_table = "obs"
[drs_settings.dialect.bar.path_specs]
dir_parts = ["foo"]
file_parts = ["foo", "bar"]
"""


def test_bad_config() -> None:
    wrong_config = Template(CONFIG).render(
        vars='{vars = "foo", attrs = "long_name", default = "__name__" }',
    )
    with pytest.raises(MetadataCrawlerException):
        DRSConfig.load(wrong_config)

    wrong_config = Template(CONFIG).render(
        vars='{var = "foo", attr = "long_name", default = "__name__" }',
        inherits="mohh",
    )
    with pytest.raises(MetadataCrawlerException):
        DRSConfig.load(wrong_config)

    wrong_config = Template(CONFIG).render(
        vars='{var = "foo", attr = "long_name", default = "__name__" }',
        fs_type="foo",
    )
    with pytest.raises(NotImplementedError):
        DRSConfig.load(wrong_config)
    wrong_config = Template(CONFIG).render(
        vars='{var = "foo", attr = "long_name", default = "__name__" }',
        sources='["foo"]',
    )
    with pytest.raises(MetadataCrawlerException):
        DRSConfig.load(wrong_config)


def test_read_zarr_data(zarr_data: Path) -> None:
    conf = Template(CONFIG).render(
        vars="{var = '{vars}', attr = 'short_name', default = '__name__' }",
        path=str(zarr_data),
    )
    cfg = toml.loads(conf)
    config = DRSConfig.load(cfg)
    data = config.read_metadata("bar", {"path": str(zarr_data), "metadata": {}})
    assert "model" in data and isinstance(data["model"], list)
    conf = Template(CONFIG).render(
        vars="{var = '{vars}', attr = 'short_name', default = '__name__' }",
        path=str(zarr_data),
        sources='["data", "path"]',
    )
    config = DRSConfig.load(conf)
    assert "path" in config.dialect["bar"].sources
    with pytest.raises(MetadataCrawlerException):
        data = config.dialect["bar"].path_specs.get_metadata_from_path(
            Path("foo/muh_bar_zup.nc")
        )
    assert (
        len(
            config.index_schema["time_aggregation"].get_time_range(
                [datetime.now()]
            )
        )
        == 2
    )


def test_get_search():
    """Test getting the right settings."""
    conf = Template(CONFIG).render(
        vars="{var = '{vars}', attr = 'short_name', default = '__name__' }",
    )
    from metadata_crawler.run import _get_search

    assert len(_get_search(DRSConfig.load(conf).datasets)) == 1
    with pytest.raises(ValueError):
        index("foo", conf)


def test_multi_config(drs_config_path: Path) -> None:
    """Test multi config."""
    glob_files = drs_config_path.parent / "*config.toml"
    conf = Template(CONFIG).render(
        vars="{var = '{vars}', attr = 'short_name', default = '__name__' }",
    )
    cfg = DRSConfig.load(glob_files, conf)
    assert "nextgems_cycle3" in cfg.datasets
    assert "cordex-benchmark-fs" in cfg.datasets
    assert "bar" in cfg.datasets


def test_benchmark_settings(drs_config_path: Path, cat_file: Path) -> None:
    """Test some benchmark settings."""
    env = os.environ.copy()
    glob_files = drs_config_path.parent / "*config.toml"
    with mock.patch.dict(os.environ, {"MDC_MAX_FILES": "5"}, clear=False):
        with mock.patch.dict(os.environ, {"MDC_SILENT": "1"}, clear=False):
            add(
                glob_files,
                drs_config_path,
                n_procs=1,
                batch_size=3,
                store=cat_file,
                catalogue_backend="jsonlines",
                data_set=["obs-fs"],
            )
    os.environ = env
    assert cat_file.exists()
    len(intake.open_catalog(cat_file).latest.read()) < 10


@pytest.mark.asyncio
async def test_posix() -> None:
    """Test posix behaviour."""
    p = PosixPath()
    _dirs = [f async for f in p.iterdir(".")]
    assert len(_dirs)
    _dirs = [f async for f in p.iterdir("/foo/bar")]
    assert not (len(_dirs))


def test_datetime_decoder_encoder() -> None:
    """Test the Datetime Encoding/Decoding."""
    now = datetime.now()
    out = json.dumps(
        {
            "now": now,
            "foo": ["bar"],
        },
        cls=DateTimeEncoder,
    )
    assert "now" in out
    assert now.isoformat() in out

    out1 = json.loads(out.encode(), cls=DateTimeDecoder)
    assert "now" in out1
    assert out1["now"] == now


def test_reading_data_attributes(
    drs_config_path: Path, dataset: xr.Dataset, zarr_data: Path
) -> None:
    """Test reading different data attributes."""
    p = PosixPath()
    cfg = DRSConfig.load(drs_config_path)

    with TemporaryDirectory() as temp_dir:
        nc_file = Path(temp_dir) / "nc_file.nc"
        dataset.to_netcdf(nc_file, engine="h5netcdf")
        assert p.read_attr("tas.grid_label", nc_file) == "gn"
        assert p.read_attr("product", nc_file) == "foo"
        assert p.read_attr("product", nc_file, engine="h5netcdf") == "foo"
        assert p.read_attr("tas.grid_label", nc_file, engine="h5netcdf") == "gn"

    assert p.read_attr("product", zarr_data) == "foo"
    assert p.read_attr("tas.grid_label", zarr_data) == "gn"
    sw = SwiftPath(**cfg.datasets["obs-swift"].storage_options)
    path = (
        "/observations/grid/CPC/CPC/cmorph/30min/atmos/30min/r1i1p1/"
        "v20210618/pr/pr_30min_CPC_cmorph_r1i1p1_201609020000-201609020030.nc"
    )
    assert sw.read_attr("ensemble", path) == "r1i1p1"



"""Test general utilities."""

from __future__ import annotations

import os
import sys
from datetime import datetime, timedelta
from typing import Optional

import mock
import pytest

from metadata_crawler.utils import IndexProgress, convert_str_to_timestamp


@pytest.mark.parametrize("interactive", [None, True, False])
@pytest.mark.parametrize("total", [10, 0])
def test_progess_bar(interactive: Optional[bool], total: int) -> None:
    """Test the interacitve progess bar."""
    max_it = 10
    with mock.patch.dict(os.environ, {"MDC_LOG_INTERVAL": "0"}, clear=True):
        p = IndexProgress(interactive=interactive, total=total)
        try:
            p.start()
            for i in range(max_it):
                p.update(1)
        finally:
            p.stop()
        assert p._done == max_it


def test_empty_string_returns_alternative() -> None:
    alt = "2000-01-01"
    dt = convert_str_to_timestamp("", alternative=alt)
    assert dt == datetime.fromisoformat(alt)


def test_non_digit_string_returns_alternative() -> None:
    alt = "1999-12-31"
    dt = convert_str_to_timestamp("fx", alternative=alt)
    assert dt == datetime.fromisoformat(alt)


def test_year_only_branch_falls_back_to_alternative() -> None:
    alt = "1980-01-01"
    dt = convert_str_to_timestamp("1999", alternative=alt)
    assert dt.year == 1999


def test_year_month_branch_falls_back_to_alternative() -> None:
    alt = "1970-01-01"
    dt = convert_str_to_timestamp("202203", alternative=alt)
    assert dt.year == 2022
    assert dt.month == 3


def test_year_dayofyear_exact_7_digits() -> None:
    dt = convert_str_to_timestamp("2022203")
    assert dt == datetime(2022, 7, 22)


def test_full_date_8_digits() -> None:
    dt = convert_str_to_timestamp("20220131")
    assert dt == datetime(2022, 1, 31)


def test_datetime_digits_only_drop_seconds() -> None:
    dt = convert_str_to_timestamp("20220131123456")
    assert dt == datetime(2022, 1, 31, 12, 34)


def test_datetime_with_T_and_minutes() -> None:
    dt = convert_str_to_timestamp("2022-01-31T1234")
    assert dt == datetime(2022, 1, 31, 12, 34)


def test_len_gt_8_without_T_and_hour_only_falls_back_to_alternative() -> None:
    alt = "1900-01-01"
    dt = convert_str_to_timestamp("2022013112", alternative=alt)
    assert dt != datetime.fromisoformat(alt)


@pytest.mark.xfail(
    sys.version_info < (3, 13),
    reason="Python <3.13 may reject 'YYYY-MM-DDT%H' in datetime.fromisoformat",
)
def test_with_T_and_hour_only_uses_fallback_to_hours() -> None:
    dt = convert_str_to_timestamp("2022-03-04T7")
    assert dt == datetime(2022, 3, 4, 7, 0)


# tests/test_convert_str_to_timestamp.py


@pytest.mark.parametrize(
    "time_str, alternative, expected",
    [
        ("", "0001-01-01", datetime.fromisoformat("0001-01-01")),
        ("fx", "0001-01-01", datetime.fromisoformat("0001-01-01")),
        ("2022", "0001-01-01", datetime.fromisoformat("2022-01-01T00:00")),
        ("202201", "1999-12-31", datetime.fromisoformat("2022-01-31T00:00")),
        ("2022203", "0001-01-01", datetime(2022, 1, 1) + timedelta(days=203 - 1)),
        ("20220101", "0001-01-01", datetime.fromisoformat("2022-01-01")),
        ("2022010112", "0001-01-01", datetime.fromisoformat("2022-01-01T12")),
        (
            "202201011234",
            "0001-01-01",
            datetime.fromisoformat("2022-01-01T12:34"),
        ),
        (
            "20220101123456",
            "0001-01-01",
            datetime.fromisoformat("2022-01-01T12:34"),
        ),
        (
            "2022-07-22T12:34",
            "0001-01-01",
            datetime.fromisoformat("2022-07-22T12:34"),
        ),
        ("2022-01-01T", "0001-01-01", datetime.fromisoformat("2022-01-01T00")),
    ],
)
def test_convert_str_to_timestamp(time_str, alternative, expected):
    assert convert_str_to_timestamp(time_str, alternative) == expected


def test_custom_alternative_used_on_failure():
    alt = "1999-12-31"
    assert convert_str_to_timestamp("nonsense", alt) == datetime.fromisoformat(
        alt
    )



"""Definitions for jinja2 templating."""

import os
from functools import lru_cache
from typing import Any, Dict, Mapping, Optional

from jinja2 import Environment, Template, Undefined

ENV = Environment(undefined=Undefined, autoescape=True)


@lru_cache(maxsize=1024)
def _compile_jinja_template(s: str) -> Template:
    return ENV.from_string(s)


class TemplateMixin:
    """Apply templating egine jinja2."""

    env_map: Optional[Dict[str, str]] = None
    _rendered = False

    def prep_template_env(self) -> None:
        """Prepare the jinja2 env."""

        def _env_get(name: str, default: Optional[str] = None) -> Optional[str]:
            return os.getenv(name, default)

        def _getenv_filter(
            varname: str, default: Optional[str] = None
        ) -> Optional[str]:
            return os.getenv(varname, default)

        ENV.globals.setdefault("env", _env_get)
        ENV.globals.setdefault("ENV", dict(os.environ))
        ENV.filters.setdefault("getenv", _getenv_filter)
        self._rendered = True

    def render_templates(
        self,
        data: Any,
        context: Mapping[str, Any],
        *,
        max_passes: int = 2,
    ) -> Any:
        """Recursively render Jinja2 templates found in strings within data.

        This function traverses common container types (``dict``, ``list``,
        ``tuple``, ``set``), dataclasses, namedtuples, and ``pathlib.Path`` objects.
        Every string encountered is treated as a Jinja2 template and rendered with
        the provided ``context``. Rendering can be repeated up to ``max_passes``
        times to resolve templates that produce further templates on the first pass.

        Parameters
        ^^^^^^^^^^
        data:
            Arbitrary Python data structure. Supported containers are ``dict``
            (keys and values), ``list``, ``tuple`` (including namedtuples),
            ``set``, dataclasses (fields), and ``pathlib.Path``.
            Scalars (e.g., ``int``, ``float``, ``bool``, ``None``) are returned
            unchanged. Strings are rendered as Jinja2 templates.
        context:
            Mapping of template variables available to Jinja2 during rendering.
        max_passes:
            Maximum number of rendering passes to perform on each string,
            by default ``2``. Increase this if templates generate further
            templates that need resolution.

        Returns
        ^^^^^^^
        Any:
            A structure of the same shape with all strings rendered. Container and
            object types are preserved where feasible (e.g., ``tuple`` stays a
            ``tuple``, namedtuple stays a namedtuple, dataclass remains the
            same dataclass type).

        Raises
        ^^^^^^^
        jinja2.TemplateError
            For other Jinja2 template errors encountered during rendering.

        Notes
        ^^^^^^
        * Dictionary keys are also rendered if they are strings (or nested
          containers with strings). If rendering causes key collisions, the
          **last** rendered key wins.
        * For dataclasses, all fields are rendered and a new instance is returned using
          ``dataclasses.replace``. Frozen dataclasses are supported.
        * Namedtuples are detected via the ``_fields`` attribute and
          reconstructed with the same type.

        Examples
        ^^^^^^^^^

            .. code-block:: python

                data = {
                    "greeting": "Hello, {{ name }}!",
                    "items": ["{{ count }} item(s)", 42],
                    "path": {"root": "/home/{{ user }}", "cfg": "{{ root }}/cfg"},
                }
                ctx = {"name": "Ada", "count": 3, "user": "ada", "root": "/opt/app"}
                TemplateMixin().render_templates(data, ctx)
                # {'greeting': 'Hello, Ada!',
                #   'items': ['3 item(s)', 42],
                #    'path': {'root': '/home/ada', 'cfg': '/opt/app/cfg'}}

        """
        if not self._rendered:
            self.prep_template_env()

        def _render_str(s: str) -> str:
            out = s
            if ("{{" not in s) and ("{%" not in s):
                return out
            for _ in range(max_passes):
                new = _compile_jinja_template(out).render(context)
                if new == out:
                    break
                out = new
            return out

        def _walk(obj: Any) -> Any:
            if isinstance(obj, str):
                return _render_str(obj)

            if isinstance(obj, dict):
                rendered: dict[Any, Any] = {}
                for k, v in obj.items():
                    rk = _render_str(k) if isinstance(k, str) else k
                    rendered[rk] = _walk(v)
                return rendered

            if isinstance(obj, list):
                return [_walk(x) for x in obj]

            if isinstance(obj, tuple):
                return tuple(_walk(x) for x in obj)

            if isinstance(obj, set):
                return {_walk(x) for x in obj}

            return obj

        return _walk(data)



"""Apply the metadata collector."""

import os
import time
from fnmatch import fnmatch
from pathlib import Path
from types import NoneType
from typing import Any, Collection, Dict, List, Optional, Sequence, Union, cast

import tomlkit
import yaml
from rich.prompt import Prompt

from .api.config import CrawlerSettings, Datasets, DRSConfig, strip_protocol
from .api.metadata_stores import (
    CatalogueBackendType,
    CatalogueReader,
    IndexName,
)
from .data_collector import DataCollector
from .logger import apply_verbosity, get_level_from_verbosity, logger
from .utils import (
    Console,
    EmptyCrawl,
    IndexProgress,
    MetadataCrawlerException,
    find_closest,
    load_plugins,
    timedelta_to_str,
)

FilesArg = Optional[Union[str, Path, Sequence[Union[str, Path]]]]


def _norm_files(catalogue_files: FilesArg) -> List[str]:
    if catalogue_files is None:
        return [""]
    return (
        [str(catalogue_files)]
        if isinstance(catalogue_files, (str, Path))
        else [str(p) for p in catalogue_files]
    )


def _match(match: str, items: Collection[str]) -> List[str]:
    out: List[str] = []
    for item in items:
        if fnmatch(item, match):
            out.append(item)

    if not out:
        msg = find_closest(f"No such dataset: {match}", match, items)
        raise MetadataCrawlerException(msg) from None
    return out


def _get_num_of_indexed_objects(
    catalogue_files: FilesArg, storage_options: Optional[Dict[str, Any]] = None
) -> int:
    num_objects = 0
    storage_options = storage_options or {}
    for cat_file in _norm_files(catalogue_files):
        try:
            cat = CatalogueReader.load_catalogue(cat_file, **storage_options)
            num_objects += cat.get("metadata", {}).get("indexed_objects", 0)
        except (FileNotFoundError, IsADirectoryError, yaml.parser.ParserError):
            pass
    return num_objects


def _get_search(
    config: Dict[str, Datasets],
    search_dirs: Optional[List[str]] = None,
    datasets: Optional[List[str]] = None,
) -> list[CrawlerSettings]:
    _search_items = []
    search_dirs = search_dirs or []
    datasets = datasets or []
    if not datasets and not search_dirs:
        return [
            CrawlerSettings(name=k, search_path=cfg.root_path)
            for (k, cfg) in config.items()
        ]
    for item in datasets or []:
        for ds in _match(item, config.keys()):
            logger.debug("Adding dataset %s", ds)
            _search_items.append(
                CrawlerSettings(name=ds, search_path=config[ds].root_path)
            )
    for num, _dir in enumerate(map(strip_protocol, search_dirs or [])):
        for name, cfg in config.items():
            if _dir.is_relative_to(strip_protocol(cfg.root_path)):
                logger.debug("Adding dataset %s", name)
                _search_items.append(
                    CrawlerSettings(name=name, search_path=str(search_dirs[num]))
                )

    return _search_items


async def async_call(
    index_system: str,
    method: str,
    batch_size: int = 2500,
    catalogue_files: Optional[Sequence[Union[Path, str]]] = None,
    verbosity: int = 0,
    log_suffix: Optional[str] = None,
    num_objects: int = 0,
    *args: Any,
    **kwargs: Any,
) -> None:
    """Add / Delete metadata from index."""
    env = cast(os._Environ[str], os.environ.copy())
    old_level = apply_verbosity(verbosity, suffix=log_suffix)

    try:
        progress = IndexProgress(total=num_objects)
        os.environ["MDC_LOG_INIT"] = "1"
        os.environ["MDC_LOG_LEVEL"] = str(get_level_from_verbosity(verbosity))
        os.environ["MDC_LOG_SUFFIX"] = (
            log_suffix or os.getenv("MDC_LOG_SUFFIX") or ""
        )
        backends = load_plugins("metadata_crawler.ingester")
        try:
            cls = backends[index_system]
        except KeyError:
            msg = find_closest(
                f"No such backend: {index_system}", index_system, backends.keys()
            )
            raise ValueError(msg) from None
        flat_files = _norm_files(catalogue_files)
        flat_files = flat_files or [""]
        storage_options = kwargs.pop("storage_options", {})
        progress.start()
        for cf in flat_files:
            async with cls(
                batch_size=batch_size,
                catalogue_file=cf or None,
                storage_options=storage_options,
                progress=progress,
            ) as obj:
                func = getattr(obj, method)
                await func(**kwargs)

    finally:
        os.environ = env
        progress.stop()
        logger.set_level(old_level)


async def async_index(
    index_system: str,
    *catalogue_files: Union[Path, str, List[str], List[Path]],
    batch_size: int = 2500,
    verbosity: int = 0,
    log_suffix: Optional[str] = None,
    **kwargs: Any,
) -> None:
    """Index metadata in the indexing system.

    Parameters
    ^^^^^^^^^^

    index_system:
        The index server where the metadata is indexed.
    catalogue_file:
        Path to the file where the metadata was stored.
    batch_size:
        If the index system supports batch-sizes, the size of the batches.
    verbosity:
        Set the verbosity of the system.
    log_suffix:
        Add a suffix to the log file output.

    Other Parameters
    ^^^^^^^^^^^^^^^^

    **kwargs:
        Keyword arguments used to delete data from the index.


    Example
    ^^^^^^^

    .. code-block:: python

        await async_index(
           "solr"
            "/tmp/catalog.yaml",
            server="localhost:8983",
            batch_size=1000,
        )
    """
    kwargs.setdefault("catalogue_files", catalogue_files)
    await async_call(
        index_system,
        "index",
        batch_size=batch_size,
        verbosity=verbosity,
        log_suffix=log_suffix,
        num_objects=_get_num_of_indexed_objects(
            kwargs["catalogue_files"],
            storage_options=kwargs.get("storage_options"),
        ),
        **kwargs,
    )


async def async_delete(
    index_system: str,
    batch_size: int = 2500,
    verbosity: int = 0,
    log_suffix: Optional[str] = None,
    **kwargs: Any,
) -> None:
    """Delete metadata from the indexing system.

    Parameters
    ^^^^^^^^^^^
    index_system:
        The index server where the metadata is indexed.
    batch_size:
        If the index system supports batch-sizes, the size of the batches.
    verbosity:
        Set the verbosity of the system.
    log_suffix:
        Add a suffix to the log file output.

    Other Parameters
    ^^^^^^^^^^^^^^^^^

    **kwargs:
        Keyword arguments used to delete data from the index.

    Examples
    ^^^^^^^^

    .. code-block:: python

        await async_delete(
            "solr"
            server="localhost:8983",
            latest_version="latest",
            facets=[("file", "*.nc"), ("project", "OBS")],
        )
    """
    await async_call(
        index_system,
        "delete",
        batch_size=batch_size,
        verbosity=verbosity,
        log_suffix=log_suffix,
        **kwargs,
    )


async def async_add(
    *config_files: Union[Path, str, Dict[str, Any], tomlkit.TOMLDocument],
    store: Optional[
        Union[str, Path, Dict[str, Any], tomlkit.TOMLDocument]
    ] = None,
    data_object: Optional[Union[str, List[str]]] = None,
    data_set: Optional[Union[List[str], str]] = None,
    data_store_prefix: str = "metadata",
    batch_size: int = 25_000,
    comp_level: int = 4,
    storage_options: Optional[Dict[str, Any]] = None,
    shadow: Optional[Union[str, List[str]]] = None,
    catalogue_backend: CatalogueBackendType = "jsonlines",
    latest_version: str = IndexName().latest,
    all_versions: str = IndexName().all,
    password: bool = False,
    n_procs: Optional[int] = None,
    verbosity: int = 0,
    log_suffix: Optional[str] = None,
    fail_under: int = -1,
    **kwargs: Any,
) -> None:
    """Harvest metadata from storage systems and add them to an intake catalogue.

    .. versionchanged:: 2511.0.0

       The catalogue argument has been rearanged and is now a keyword
       argument: ``async_add("data.yaml", "drs-config.toml")`` becomes
       ``async_add("drs-config.toml", store="data.yaml")``. If the ``store`` keyword
       is omitted the output catalogue will be interpreted as config file.

    Parameters
    ^^^^^^^^^^

    config_files:
        Path to the drs-config file / loaded configuration.
    store:
        Path to the intake catalogue.
    data_objects:
        Instead of defining datasets that are to be crawled you can crawl
        data based on their directories. The directories must be a root dirs
        given in the drs-config file. By default all root dirs are crawled.
    data_object:
        Objects (directories or catalogue files) that are processed.
    data_set:
        Dataset(s) that should be crawled. The datasets need to be defined
        in the drs-config file. By default all datasets are crawled.
        Names can contain wildcards such as ``xces-*``.
    data_store_prefix: str
        Absolute path or relative path to intake catalogue source
    batch_size:
        Batch size that is used to collect the meta data. This can affect
        performance.
    comp_level:
        Compression level used to write the meta data to csv.gz
    storage_options:
        Set additional storage options for adding metadata to the metadata store
    shadow:
        'Shadow' this storage options. This is useful to hide secrets in public
        data catalogues.
    catalogue_backend:
        Intake catalogue backend
    latest_version:
        Name of the core holding 'latest' metadata.
    all_versions:
        Name of the core holding 'all' metadata versions.
    password:
        Display a password prompt before beginning
    n_procs:
        Set the number of parallel processes for collecting.
    verbosity:
        Set the verbosity of the system.
    log_suffix:
        Add a suffix to the log file output.
    fail_under:
        Fail if less than X of the discovered files could be indexed.

    Other Parameters
    ^^^^^^^^^^^^^^^^

    **kwargs:
        Additional keyword arguments.


    Examples
    ^^^^^^^^

     .. code-block:: python

        await async_add(
             "~/data/drs-config.toml",
             store="my-data.yaml",
             data_set=["cmip6", "cordex"],
        )

    """
    env = cast(os._Environ[str], os.environ.copy())
    old_level = apply_verbosity(verbosity, suffix=log_suffix)
    try:
        os.environ["MDC_LOG_INIT"] = "1"
        os.environ["MDC_LOG_LEVEL"] = str(get_level_from_verbosity(verbosity))
        os.environ["MDC_LOG_SUFFIX"] = (
            log_suffix or os.getenv("MDC_LOG_SUFFIX") or ""
        )
        config_files = config_files or (
            os.environ.get("EVALUATION_SYSTEM_CONFIG_DIR", ""),
        )
        if not all(config_files):
            raise MetadataCrawlerException(
                "You must give a config file/directory"
            )
        st = time.time()
        passwd: Optional[str] = None
        if password:  # pragma: no cover
            passwd = Prompt.ask(
                "[b]Enter the password", password=True
            )  # pragma: no cover

        if passwd:
            os.environ["DRS_STORAGE_PASSWD"] = passwd
        data_object = (
            data_object
            if isinstance(data_object, (NoneType, list))
            else [str(data_object)]
        )
        data_set = (
            data_set
            if isinstance(data_set, (NoneType, list))
            else [str(data_set)]
        )
        cfg = DRSConfig.load(*config_files)
        async with DataCollector(
            cfg,
            store,
            IndexName(latest=latest_version, all=all_versions),
            *_get_search(cfg.datasets, data_object, data_set),
            batch_size=batch_size,
            comp_level=comp_level,
            backend=catalogue_backend,
            data_store_prefix=data_store_prefix,
            n_procs=n_procs,
            storage_options=storage_options or {},
            shadow=shadow,
            **kwargs,
        ) as data_col:
            await data_col.ingest_data()
            num_files = data_col.ingested_objects
            files_discovered = data_col.crawled_files
            dt = timedelta_to_str(time.time() - st)
        logger.info("Discovered: %s files", f"{files_discovered:10,.0f}")
        logger.info("Ingested: %s files", f"{num_files:10,.0f}")
        logger.info("Spend: %s", dt)
        Console.print(" " * Console.width, end="\r")
        Console.print(
            (
                f"[bold]Ingested [green]{num_files:10,.0f}[/green] "
                f"within [green]{dt}[/green][/bold]"
            )
        )

        if (
            files_discovered >= fail_under and num_files < fail_under
        ) or files_discovered == 0:
            if data_col.ingest_queue.silent is False:
                await data_col.ingest_queue.delete()
                raise EmptyCrawl(
                    "Could not fulfill discovery threshold!"
                ) from None
    finally:
        os.environ = env
        logger.set_level(old_level)



"""Definitions for path manipulatins."""

from pathlib import Path
from typing import Tuple, Union
from urllib.parse import urlsplit

import fsspec


class PathMixin:
    """Class that defines typical Path operations."""

    async def suffix(self, path: Union[str, Path]) -> str:
        """Get the suffix of a given input path.

        Parameters
        ^^^^^^^^^^
        path: str, asyncio.Path, pathlib.Path
            Path of the object store

        Returns
        ^^^^^^-
        str: The file type extension of the path.
        """
        return Path(path).suffix

    def get_fs_and_path(self, uri: str) -> Tuple[fsspec.AbstractFileSystem, str]:
        """Return (fs, path) suitable for xarray.

        Parameters
        ^^^^^^^^^^
        uri:
            Path to the object store / file name


        Returns
        ^^^^^^-
        fsspec.AbstractFileSystem, str:
            The AbstractFileSystem class and the corresponding path to the
            data store.
        """
        protocol, path = fsspec.core.split_protocol(uri)
        protocol = protocol or "file"
        path = urlsplit(uri.removeprefix(f"{protocol}://")).path
        return fsspec.filesystem(protocol), path



from pathlib import Path
from typing import Any

import tomlkit

from ._version import __version__ as __version__
from .api.config import ConfigMerger
from .api.metadata_stores import CatalogueBackendType
from .data_collector import DataCollector as DataCollector
from .logger import logger as logger
from .run import async_add as async_add
from .run import async_delete as async_delete
from .run import async_index as async_index

__all__ = [
    "logger",
    "__version__",
    "DataCollector",
    "index",
    "add",
    "delete",
    "get_config",
    "async_index",
    "async_delete",
    "async_add",
    "get_config",
]

def get_config(config: Path | str | None = None) -> ConfigMerger: ...
def index(
    index_system: str,
    *catalogue_files: Path | str | list[str] | list[Path],
    batch_size: int = 2500,
    verbosity: int = 0,
    log_suffix: str | None = None,
    **kwargs: Any,
) -> None: ...
def delete(
    index_system: str,
    batch_size: int = 2500,
    verbosity: int = 0,
    log_suffix: str | None = None,
    **kwargs: Any,
) -> None: ...
def add(
    store: str | Path | None = None,
    config_file: Path | str | dict[str, Any] | tomlkit.TOMLDocument | None = None,
    data_object: str | list[str] | None = None,
    data_set: str | list[str] | None = None,
    data_store_prefix: str = "metadata",
    catalogue_backend: CatalogueBackendType = "jsonlines",
    batch_size: int = 25000,
    comp_level: int = 4,
    storage_options: dict[str, Any] | None = None,
    shadow: str | list[str] | None = None,
    latest_version: str = ...,
    all_versions: str = ...,
    n_procs: int | None = None,
    verbosity: int = 0,
    log_suffix: str | None = None,
    password: bool = False,
    fail_under: int = -1,
    **kwargs: Any,
) -> None: ...



from typing import Awaitable, List, Optional, Sequence

class posix:
    @staticmethod
    def rglob(
        path: str,
        glob_pattern: Optional[str] = ...,
        suffixes: Optional[List[str]] = ...,
    ) -> Awaitable[Sequence[str]]: ...
    @staticmethod
    def is_file(path: str) -> Awaitable[bool]: ...
    @staticmethod
    def is_dir(path: str) -> Awaitable[bool]: ...
    @staticmethod
    def iterdir(path: str) -> Awaitable[Sequence[str]]: ...



"""Test crawling intake catalogues."""

import json
import os
from pathlib import Path
from tempfile import NamedTemporaryFile

import intake
import mock
import numpy as np
import pandas as pd
import pytest

from metadata_crawler import add
from metadata_crawler.backends.intake import IntakePath
from metadata_crawler.cli import walk_catalogue


def test_crawl_intake_catalogue(
    drs_config_path: Path,
    cat_file: Path,
    data_dir: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test if we can crawl intake catalogues."""
    monkeypatch.chdir(data_dir.parent)
    with mock.patch.dict(os.environ, {"MDC_INTERACTIVE": "1"}, clear=True):
        inp = data_dir / "intake" / "catalog" / "catalog.yaml"
        add(
            drs_config_path,
            store=cat_file,
            data_object=[inp],
            verbosity=10,
            batch_size=3,
            n_procs=1,
            catalogue_backend="jsonlines",
        )
    assert cat_file.exists()
    cat = intake.open_catalog(cat_file)
    assert len(cat.latest.read()) > 0


def test_walk_intake_catalogue(data_dir: Path) -> None:
    """Test walking the catalogue."""
    cat_file = data_dir / "intake" / "catalog" / "dkrz_dyamond-winter_disk.json"
    num = walk_catalogue(cat_file)
    assert num > 100


def test_intake_utils() -> None:
    p = IntakePath()
    with NamedTemporaryFile(suffix=".json") as temp_d:
        Path(temp_d.name).write_text(json.dumps({"foo": "bar"}))
        assert p._is_esm_catalogue(temp_d.name) is False
        Path(temp_d.name).write_text(json.dumps({"esmcat": "bar"}))
        assert p._is_esm_catalogue(temp_d.name) is True
        lines = "first" + 30 * "foo\n" + "esmcat"
        Path(temp_d.name).write_text(lines)
        assert p._is_esm_catalogue(temp_d.name) is False
    with NamedTemporaryFile(suffix=".foo") as temp_d:
        Path(temp_d.name).write_text(json.dumps({"esmcat": "bar"}))
        assert p._is_esm_catalogue(temp_d.name) is False
    df = pd.DataFrame(
        {
            "array": [np.array([1, 2])],
            "nan": [pd.NaT],
            "string": ["foo"],
            "int": [3],
            "list": [[1, 2]],
        }
    )
    assert isinstance(p._to_py(p), IntakePath)
    assert p._to_py(df["nan"][0]) is None
    assert isinstance(p._to_py(df["array"][0]), list)
    assert isinstance(p._to_py(df["string"][0]), str)
    assert isinstance(p._to_py(df["int"][0]), int)
    assert isinstance(p._to_py(df["list"][0]), list)



"""Definitions for lookup table mixins."""

import atexit
import os
from types import MappingProxyType
from typing import Any, Dict, Mapping, Tuple

from appdirs import user_cache_dir
from diskcache import Cache

from .lookup_tables import cmor_lookup as _NESTED

Key = Tuple[str, ...]


def _flatten_static(
    prefix: Tuple[str, ...], node: Mapping[str, Any], out: Dict[Key, Any]
) -> None:
    """Flatten nested CMOR-like dict.

      cmip6 -> CF3hr -> tas -> {'realm': 'atmos', 'time-frequency': '3hrPt', ...}
    into keys ('cmip6','CF3hr','tas','realm') -> 'atmos'.
    """
    for k, v in node.items():
        if isinstance(v, Mapping):
            if v and not all(isinstance(x, Mapping) for x in v.values()):
                for leaf_k, leaf_v in v.items():
                    out[prefix + (k, leaf_k)] = leaf_v
            else:
                _flatten_static(prefix + (k,), v, out)
        else:
            out[prefix + (k,)] = v


_flat: Dict[Key, Any] = {}

_dir = os.getenv("MDC_LOOKUP_CACHE_DIR") or os.path.join(
    user_cache_dir("metadata-crawler", "freva"), "lookup"
)
os.makedirs(_dir, exist_ok=True)
_DC = Cache(
    _dir, size_limit=2 * 1024**3, eviction="least-recently-used", cull_limit=10
)
atexit.register(_DC.close)


class LookupMixin:
    """Provide a Mixing with a process safe lookup().

    The mixin does:
      - process-wide static table (CMOR) via CMOR_STATIC
      - per-instance disk cache for file-derived attrs
      - in-flight de-duplication for concurrent misses

    Subclass must implement:
      def read_attr(self, attribute: str, path: str, **read_kws: Any) -> Any
    """

    CMOR_STATIC: Mapping[Key, Any] = {}

    def set_static_from_nested(self) -> None:
        """Flatting the cmor lookup table."""
        if not self.CMOR_STATIC:
            _flatten_static((), _NESTED, _flat)
            self.CMOR_STATIC = MappingProxyType(_flat)

    def read_attr(self, attribute: str, path: str, **read_kws: Any) -> Any:
        """Get a metadata attribute from a datastore object."""
        raise NotImplementedError  # pragma: no cover

    def lookup(
        self, path: str, attribute: str, *tree: str, **read_kws: Any
    ) -> Any:
        """Get metadata from a lookup table.

        This function will read metadata from a pre-defined cache table and if
        the metadata is not present in the cache table it'll read the
        the object store and add the metadata to the cache table.

        Parameters
        ^^^^^^^^^^

        path:
            Path to the object store / file name
        attribute:
            The attribute that is retrieved from the data.
            variable attributes can be defined by a ``.``.
            For example: ``tas.long_name`` would get attribute ``long_name``
            from variable ``tas``.
        *tree:
            A tuple representing nested attributes. Attributes are nested for
            more efficient lookup. ('atmos', '1hr', 'tas') will translate into
            a tree of ['atmos']['1hr']['tas']

        Other Parameters
        ^^^^^^^^^^^^^^^^
        **read_kws:
            Keyword arguments passed to open the datasets.

        """
        # 1) static fast-path
        try:
            return self.CMOR_STATIC[tree]
        except KeyError:
            pass
        # 2) process-safe disk cache (key includes path)
        val = _DC.get(tree)
        if val is None:
            val = self.read_attr(attribute, path, **read_kws)
            if not _DC.add(tree, val):
                val = _DC.get(tree)
        return val



"""API for adding new cataloging systems."""

from __future__ import annotations

import abc
from pathlib import Path
from types import TracebackType
from typing import (
    Any,
    AsyncIterator,
    Dict,
    List,
    Optional,
    Self,
    Tuple,
    Type,
    Union,
    cast,
)

from ..logger import logger
from ..utils import Console, IndexProgress
from .config import SchemaField
from .metadata_stores import CatalogueReader, IndexStore


class BaseIndex:
    """Base class to index metadata in the indexing system.

    Any data ingestion class that implements metadata ingestion into
    cataloguing systems should inherit from this class.

    This abstract class will setup consumer threads and a fifo queue that wait
    for new data to harvest metadata and add it to the cataloguing system.
    Only :py:func:`add` and :py:func:`delete` are abstract methods that need
    to be implemented for each cataloguing ingestion class. The rest is done
    by this base class.

    Parameters
    ^^^^^^^^^^
    catalogue_file:
        Path to the intake catalogue
    batch_size:
        The amount for metadata that should be gathered `before` ingesting
        it into the catalogue.
    progress:
        Optional rich progress object that should display the progress of the
        tasks.

    Attributes
    ^^^^^^^^^^
    """

    def __init__(
        self,
        catalogue_file: Optional[Union[str, Path]] = None,
        batch_size: int = 2500,
        storage_options: Optional[Dict[str, Any]] = None,
        progress: Optional[IndexProgress] = None,
        **kwargs: Any,
    ) -> None:
        self._store: Optional[IndexStore] = None
        self.progress = progress or IndexProgress(total=-1)
        if catalogue_file is not None:
            _reader = CatalogueReader(
                catalogue_file=catalogue_file or "",
                batch_size=batch_size,
                storage_options=storage_options,
            )
            self._store = _reader.store
        self.__post_init__()

    def __post_init__(self) -> None: ...

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None: ...

    @property
    def index_schema(self) -> Dict[str, SchemaField]:
        """Get the index schema."""
        return cast(Dict[str, SchemaField], getattr(self._store, "schema", {}))

    @property
    def index_names(self) -> Tuple[str, str]:
        """Get the names of the indexes for latests and all data."""
        return cast(
            Tuple[str, str], getattr(self._store, "index_names", ("", ""))
        )

    async def get_metadata(
        self, index_name: str
    ) -> AsyncIterator[List[Dict[str, Any]]]:
        """Get the metadata of an index in batches.

        Parameters
        ^^^^^^^^^^
        index_name:
            Name of the index that should be read.
        """
        if self._store:
            batch = []
            num_items = 0
            logger.info("Indexing %s", index_name)
            async for batch in self._store.read(index_name):
                yield batch
                self.progress.update(len(batch))
                num_items += len(batch)
            msg = f"Indexed {num_items:10,.0f} items for index {index_name}"
            Console.print(msg) if Console.is_terminal else print(msg)

    @abc.abstractmethod
    async def delete(self, **kwargs: Any) -> None:
        """Delete data from the cataloguing system.

        Parameters
        ^^^^^^^^^^
        flush:
            Boolean indicating whether or not the data should be flushed after
            amending the catalogue (if implemented).
        search_keys:
            key-value based query for data that should be deleted.
        """

    @abc.abstractmethod
    async def index(
        self,
        metadata: Optional[dict[str, Any]] = None,
        core: Optional[str] = None,
        **kwags: Any,
    ) -> None:
        """Add metadata into the cataloguing system.

        Parameters
        ^^^^^^^^^^
        metadata_batch:
            batch of metadata stored in a two valued tuple. The first entry
            of the tuple represents a name of the catalog. This entry
            might have different meanings for different cataloguing systems.
            For example apache solr will receive the name of the ``core``.
            The second  entry is the meta data itself, saved in a dictionary.
        flush:
            Boolean indicating whether or not the data should be flushed after
            adding to the catalogue (if implemented)
        """



"""Test of cftime_utils."""

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from metadata_crawler.utils.cftime_utils import (
    infer_cmor_like_time_frequency,
)


def _make_ds_from_times(
    times: pd.DatetimeIndex | list,
    coord_name: str = "time",
    dim_name: str | None = None,
    attrs: dict | None = None,
) -> xr.Dataset:
    """
    Helper: create a minimal Dataset with a given time-like coordinate.
    - coord_name: name of the coordinate variable
    - dim_name:   underlying dimension name (defaults to coord_name)
    """
    if dim_name is None:
        dim_name = coord_name

    time_da = xr.DataArray(
        times,
        dims=(dim_name,),
        name=coord_name,
        attrs=attrs or {},
    )
    data = xr.DataArray(
        np.arange(len(time_da)),
        dims=(dim_name,),
        name="var",
    )

    ds = xr.Dataset({"var": data, coord_name: time_da})
    ds = ds.set_coords(coord_name)
    ds.coords[coord_name].attrs = attrs or {}
    return ds


# ----------------------------------------------------------------------
# Edge cases: missing / insufficient / constant time
# ----------------------------------------------------------------------


def test_no_time_coord_returns_fx() -> None:
    ds = xr.Dataset({"var": ("x", [1, 2, 3])})
    freq = infer_cmor_like_time_frequency(ds)
    assert freq == "fx"


def test_single_timestamp_returns_fx() -> None:
    times = pd.date_range("2000-01-01", periods=1, freq="D")
    ds = _make_ds_from_times(times)
    freq = infer_cmor_like_time_frequency(ds)
    assert freq == "fx"


def test_constant_time_returns_fx() -> None:
    times = [pd.Timestamp("2000-01-01")] * 5
    ds = _make_ds_from_times(times)
    freq = infer_cmor_like_time_frequency(ds)
    assert freq == "fx"


# ----------------------------------------------------------------------
# Sub-hourly and hourly frequencies
# ----------------------------------------------------------------------


def test_subhourly_mapped_to_subhr() -> None:
    times = pd.date_range("2000-01-01", periods=4, freq="15min")
    ds = _make_ds_from_times(times)
    freq = infer_cmor_like_time_frequency(ds)
    assert freq == "subhr"


@pytest.mark.parametrize(
    "freq_str, expected",
    [
        ("1h", "1hr"),
        ("3h", "3hr"),
        ("6h", "6hr"),
    ],
)
def test_hourly_like_frequencies(freq_str: str, expected: str) -> None:
    times = pd.date_range("2000-01-01", periods=4, freq=freq_str)
    ds = _make_ds_from_times(times)
    freq = infer_cmor_like_time_frequency(ds)
    assert freq == expected


# ----------------------------------------------------------------------
# Daily & multi-day frequencies
# ----------------------------------------------------------------------


def test_daily_frequency_day() -> None:
    times = pd.date_range("2000-01-01", periods=4, freq="D")
    ds = _make_ds_from_times(times)
    freq = infer_cmor_like_time_frequency(ds)
    assert freq == "day"


def test_six_day_frequency_6d() -> None:
    times = pd.date_range("2000-01-01", periods=4, freq="6D")
    ds = _make_ds_from_times(times)
    freq = infer_cmor_like_time_frequency(ds)
    assert freq == "6d"


def test_weekly_frequency_1w() -> None:
    times = pd.date_range("2000-01-01", periods=4, freq="7D")
    ds = _make_ds_from_times(times)
    freq = infer_cmor_like_time_frequency(ds)
    assert freq == "1w"


def test_biweekly_frequency_sem() -> None:
    times = pd.date_range("2000-01-01", periods=4, freq="14D")
    ds = _make_ds_from_times(times)
    freq = infer_cmor_like_time_frequency(ds)
    assert freq == "sem"


# ----------------------------------------------------------------------
# Monthly / seasonal / yearly / decadal (approximate ranges)
# ----------------------------------------------------------------------


def test_monthly_like_frequency_mon() -> None:
    # 2000-01-01 to 2000-01-31 → 30 days, inside [20, 40] day window
    times = [pd.Timestamp("2000-01-01"), pd.Timestamp("2000-01-31")]
    ds = _make_ds_from_times(times)
    freq = infer_cmor_like_time_frequency(ds)
    assert freq == "mon"


def test_seasonal_like_frequency_season() -> None:
    # ~90 days spacing
    times = [pd.Timestamp("2000-01-01"), pd.Timestamp("2000-04-01")]
    ds = _make_ds_from_times(times)
    freq = infer_cmor_like_time_frequency(ds)
    assert freq == "season"


def test_yearly_like_frequency_yr() -> None:
    times = [pd.Timestamp("2000-01-01"), pd.Timestamp("2001-01-01")]
    ds = _make_ds_from_times(times)
    freq = infer_cmor_like_time_frequency(ds)
    assert freq == "yr"


def test_decadal_like_frequency_dec() -> None:
    times = [pd.Timestamp("2000-01-01"), pd.Timestamp("2010-01-01")]
    ds = _make_ds_from_times(times)
    freq = infer_cmor_like_time_frequency(ds)
    assert freq == "dec"


# ----------------------------------------------------------------------
# Fallback behavior for non-standard steps
# ----------------------------------------------------------------------


def test_nonstandard_step_returns_generic_days_string() -> None:
    # 2-day step is not explicitly mapped; expect "<Xd>" style.
    times = [pd.Timestamp("2000-01-01"), pd.Timestamp("2000-01-03")]
    ds = _make_ds_from_times(times)
    freq = infer_cmor_like_time_frequency(ds)

    # We only assert it's *some* day-like fallback, not 'fx' and not a known code:
    assert freq not in {"fx", "day", "6d", "1w", "sem"}
    assert freq.endswith("d")


# ----------------------------------------------------------------------
# Coordinate detection logic
# ----------------------------------------------------------------------


def test_explicit_time_coord_name() -> None:
    times = pd.date_range("2000-01-01", periods=4, freq="D")
    ds = _make_ds_from_times(times, coord_name="mytime")
    freq = infer_cmor_like_time_frequency(ds, time_coord="mytime")
    assert freq == "day"


def test_detect_time_by_standard_name_attr() -> None:
    times = pd.date_range("2000-01-01", periods=4, freq="6h")
    ds = _make_ds_from_times(
        times,
        coord_name="t",
        attrs={"standard_name": "time"},
    )
    freq = infer_cmor_like_time_frequency(ds)
    assert freq == "6hr"


def test_detect_time_by_axis_T_attr() -> None:
    times = pd.date_range("2000-01-01", periods=4, freq="3h")
    ds = _make_ds_from_times(
        times,
        coord_name="t",
        attrs={"axis": "T"},
    )
    freq = infer_cmor_like_time_frequency(ds)
    assert freq == "3hr"


def test_detect_time_by_dim_named_time() -> None:
    # Dimension is 'time', coord name is different.
    times = pd.date_range("2000-01-01", periods=4, freq="D")
    time_da = xr.DataArray(times, dims=("time",), name="not_time")
    data = xr.DataArray(np.arange(len(time_da)), dims=("time",), name="var")
    ds = xr.Dataset({"not_time": time_da, "var": data}).set_coords("not_time")

    freq = infer_cmor_like_time_frequency(ds)

    # This asserts the robust behavior: using dim 'time' to deduce it.
    assert freq == "day"


# ----------------------------------------------------------------------
# Optional: cftime support (kept loose)
# ----------------------------------------------------------------------


def test_cftime_timestamps_if_supported() -> None:
    cftime = pytest.importorskip("cftime")

    times = [
        cftime.DatetimeNoLeap(2000, 1, 1),
        cftime.DatetimeNoLeap(2000, 1, 2),
        cftime.DatetimeNoLeap(2000, 1, 3),
    ]
    ds = _make_ds_from_times(times)

    freq = infer_cmor_like_time_frequency(ds)

    # Keep expectation loose depending on your implementation.
    # If you support cftime cleanly, this should be "day".
    # If not, adjust your function or relax this assertion.
    assert freq in {"day", "fx", "unknown"}



"""Interact with the INTAKE metadata catalogues."""

from __future__ import annotations

import pathlib
from fnmatch import fnmatch
from types import NoneType
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Union,
)
from urllib.parse import unquote, urlparse

import fsspec
import intake
import pandas as pd

from ..api.storage_backend import Metadata, MetadataType, PathTemplate
from ..logger import logger


class IntakePath(PathTemplate):
    """Class to interact with the Intake metadata catalogues."""

    _fs_type = None

    async def is_file(self, path: Union[str, pathlib.Path]) -> bool:
        """Check if a given path is a file."""
        return True

    async def is_dir(self, path: Union[str, pathlib.Path]) -> bool:
        """Check if a given path is a directory."""
        return False

    @staticmethod
    def _normalize_path(path: str) -> str:
        """Turn file:// URLs into OS paths; leave others as-is."""
        if isinstance(path, str) and path.startswith("file://"):
            return unquote(urlparse(path).path)
        return path

    async def _walk_yaml_catalogue(
        self,
        cat: intake.catalog.Catalog,
    ) -> AsyncIterator[MetadataType]:

        for name in cat:
            entry = cat[name]
            container = getattr(entry, "container", None)

            if container == "catalog":
                async for md in self._walk_yaml_catalogue(entry()):
                    yield md
                continue

            src = entry()
            meta = getattr(src, "_entry", src).describe() or {}
            args = meta.get("args", {})
            urlpath = (
                args.get("urlpath")
                or args.get("path")
                or args.get("url")
                or meta.get("uri")
                or meta.get("file")
                or args.get("urlpaths")
            ) or []
            for raw_path in urlpath if isinstance(urlpath, list) else [urlpath]:
                path = self._normalize_path(raw_path)
                logger.debug("Found file %s", path)
                yield MetadataType(
                    path=path,
                    metadata=getattr(src, "metadata", meta.get("metadata", {})),
                )

    @staticmethod
    def _to_py(value: Any) -> Any:
        if isinstance(value, (float, int, bool, str, NoneType)):
            return value
        try:
            if hasattr(value, "tolist"):
                return value.tolist()
            if pd.isna(value):
                return None
        except Exception:
            pass
        return value

    async def _walk_esm_catalogue(
        self,
        cat: intake.catalog.Catalog,
    ) -> AsyncIterator[MetadataType]:
        df: pd.DataFrame = getattr(cat, "df", pd.DataFrame())
        cols = list(df.columns)
        for row in df.itertuples(index=False, name=None):
            meta: Dict[str, Any] = {k: self._to_py(v) for k, v in zip(cols, row)}
            urlpath = (
                meta.get("urlpath")
                or meta.get("path")
                or meta.get("url")
                or meta.get("uri")
                or meta.get("file")
                or meta.get("urlpaths")
            ) or []
            for raw_path in urlpath if isinstance(urlpath, list) else [urlpath]:
                path = self._normalize_path(raw_path)
                logger.debug("Found file %s", path)
                yield MetadataType(path=path, metadata=meta)

    async def iterdir(
        self,
        path: Union[str, pathlib.Path],
    ) -> AsyncIterator[str]:
        """Get all sub directories from a given path.

        Parameter
        ---------
        path : str, pathlib.Path
            Path of the object store

        Yields
        ------
        str:
            1st level sub directory
        """
        yield str(path)

    def _is_esm_catalogue(self, path: str) -> bool:
        if not self._normalize_path(path).endswith(".json"):
            return False
        esmcat = False
        fs = fsspec.get_filesystem_class(
            fsspec.core.split_protocol(path)[0] or "file"
        )(**self.storage_options)
        with fs.open(path, mode="rb", **self.storage_options) as stream:
            num = 0
            for line in stream:
                if "esmcat" in line.decode("utf-8"):
                    esmcat = True
                    break
                if num > 19:
                    break
                num += 1
        return esmcat

    async def rglob(
        self, path: Union[str, pathlib.Path], glob_pattern: str = "*"
    ) -> AsyncIterator[MetadataType]:
        """Go through catalogue path."""
        path = str(path)
        if self._is_esm_catalogue(path):
            cat: intake.catalog.Catalog = intake.open_esm_datastore(
                path, **self.storage_options
            )
            func: Callable[[str], AsyncIterator[MetadataType]] = (
                self._walk_esm_catalogue
            )
        else:
            cat = intake.open_catalog(path, **self.storage_options)
            func = self._walk_yaml_catalogue

        async for md in func(cat):
            if "." + md["path"].rpartition(".")[-1] in self.suffixes and fnmatch(
                md["path"], glob_pattern
            ):
                yield md

    def path(self, path: Union[str, pathlib.Path]) -> str:
        """Get the full path (including any schemas/netlocs).

        Parameters
        ----------
        path: str, pathlib.Path
            Path of the object store

        Returns
        -------
        str:
            URI of the object store
        """
        return str(path)

    def uri(self, path: Union[str, pathlib.Path]) -> str:
        """Get the uri of the object store.

        Parameters
        ----------
        path: str, pathlib.Path
            Path of the object store

        Returns
        -------
        str:
            URI of the object store
        """
        fs_type, path = fsspec.core.split_protocol(str(path))
        fs_type = fs_type or "file"
        return f"{fs_type}://{path}"

    def fs_type(self, path: Union[str, pathlib.Path]) -> str:
        """Define the file system type."""
        fs_type, _ = fsspec.core.split_protocol(str(path))
        return fs_type or "posix"

    async def walk(self, path: str) -> AsyncIterator[Metadata]:
        """Walk a catalogue."""
        async for md in self.rglob(path):
            yield Metadata(path=md["path"], metadata=md["metadata"])
